#include <SDL.h>
#include <SDL_vulkan.h>

#include <iostream>
#include <fstream>
#include <sstream>
#include <json.hpp>

#ifdef DEBUG
#include <imgui.h>
#endif

#include "renderer.h"
#include "platform.h"
#include "world.h"
#include "mesh.h"
#include "animationsystem.h"
#include "material.h"
#include "config.h"
#include "stringutils.h"
#include "worldmanager.h"
#include "assetmanager.h"
#include "audiosystem.h"

SDL_Window* window = nullptr;
Renderer* renderer = nullptr;

std::vector<const char*> platform::getRequiredExtensions() {
    uint32_t count = 0;
    SDL_Vulkan_GetInstanceExtensions(window, &count, nullptr);

    std::vector<const char*> names(count);
    SDL_Vulkan_GetInstanceExtensions(window, &count, names.data());

    return names;
}

uint32_t platform::getTime() {
    return SDL_GetTicks();
}

int windowX = SDL_WINDOWPOS_CENTERED;
int windowY = SDL_WINDOWPOS_CENTERED;
int windowWidth = 1280;
int windowHeight = 720;
int windowFullscreen = 0;
std::string currentGraphicsPreset = "Medium";
GraphicsConfig graphicsConfig;

int toInt(const std::string &str) {
    std::stringstream ss(str);

    int num;
    if((ss >> num).fail())
        return -1;

    return num;
}

void loadGraphicsConfig() {
    Config config;
    if(!config.load("data/graphics_presets.cfg"))
        return;

    graphicsConfig.shadowResolution = toInt(config.get(currentGraphicsPreset.c_str(), "shadowResolution"));
    graphicsConfig.dofDownscale = toInt(config.get(currentGraphicsPreset.c_str(), "dofDownscale"));
    graphicsConfig.filterPCF = toInt(config.get(currentGraphicsPreset.c_str(), "filterPCF"));
}

void readConfig() {
    Config config;
    if(!config.load("user.cfg")) {
        loadGraphicsConfig(); // load medium anyway
        return;
    }

    windowX = toInt(config.get("Window", "x"));
    windowY = toInt(config.get("Window", "y"));
    windowWidth = toInt(config.get("Window", "width"));
    windowHeight = toInt(config.get("Window", "height"));
    windowFullscreen = toInt(config.get("Window", "fullscreen"));

    currentGraphicsPreset = config.get("Graphics", "preset");
    graphicsConfig.vsync = toInt(config.get("Graphics", "vsync"));

    loadGraphicsConfig();
}

void writeConfig() {
    Config config;

    config.set("Window", "x", std::to_string(windowX));
    config.set("Window", "y", std::to_string(windowY));
    config.set("Window", "width", std::to_string(windowWidth));
    config.set("Window", "height", std::to_string(windowHeight));
    config.set("Window", "fullscreen", std::to_string(windowFullscreen));

    config.set("Graphics", "preset", currentGraphicsPreset);
    config.set("Graphics", "vsync", std::to_string(graphicsConfig.vsync));

    config.save("user.cfg");
}

Cinematic* cinematic = nullptr;

int main(int argc, char* argv[]) {
    readConfig();

    bool cinematicMode = false;
    if(argc > 2 && strcmp(argv[1], "--cinematic") == 0)
        cinematicMode = true;

    bool record = false;
    if(argc > 3 && strcmp(argv[2], "--record") == 0)
        record = true;
    
    window = SDL_CreateWindow("ITER",
                              windowX,
                              windowY,
                              windowWidth,
                              windowHeight,
                              SDL_WINDOW_VULKAN |
                              SDL_WINDOW_RESIZABLE);
    if(window == nullptr)
        return -1;

    SDL_SetWindowFullscreen(window, windowFullscreen == 1 ? SDL_WINDOW_FULLSCREEN_DESKTOP : 0);

#ifdef DEBUG
    ImGui::CreateContext();

    auto& io = ImGui::GetIO();
    io.DisplaySize = ImVec2(windowWidth, windowHeight);

    ImGui::StyleColorsDark();
#endif

    renderer = new Renderer(graphicsConfig);

    VkSurfaceKHR surface = nullptr;
    SDL_Vulkan_CreateSurface(window, renderer->getInstance(), &surface);

    RenderTarget* target = renderer->createSurfaceRenderTarget(surface);

    auto audio = new AudioSystem();
    
    assetManager.setRenderer(renderer);
    assetManager.setAudioSystem(audio);

    AnimationSystem* animationSystem = new AnimationSystem();

    TransformComponent* playerTransform = nullptr, *cameraTransform = nullptr;
    CameraComponent* cameraComponent = nullptr;

    if(cinematicMode)
        cinematic = animationSystem->loadCinematic(argv[2]);
    else {
        worldManager.loadWorld("test.world");
        worldManager.switchWorld("test.world");

        EntityID entity = ECS::createEntity(worldManager.getCurrentWorld());

        cameraTransform = ECS::addComponent<TransformComponent>(entity);
        cameraTransform->position = glm::vec3(5, 5, 5);

        cameraComponent = ECS::addComponent<CameraComponent>(entity);

        EntityID playerEntity = ECS::createEntity(worldManager.getCurrentWorld());

        playerTransform = ECS::addComponent<TransformComponent>(playerEntity);
        playerTransform->position.y = 2;

        MeshComponent* mesh = ECS::addComponent<MeshComponent>(playerEntity);
        mesh->mesh = assetManager.loadMesh("player.obj");
        mesh->material = assetManager.loadMaterial("test.material");
        
        ListenerComponent* listener = ECS::addComponent<ListenerComponent>(playerEntity);
    }

    float currentTime = 0.0f, lastTime = 0.0f;

    bool running = true;
    while(running) {
        currentTime = SDL_GetTicks() / 1000.0f;
        const float deltaTime = currentTime - lastTime;
        lastTime = currentTime;

        SDL_Event event = {};
        while(SDL_PollEvent(&event) > 0) {
            if(event.type == SDL_QUIT)
                running = false;

            if(event.type == SDL_WINDOWEVENT) {
                if(event.window.event == SDL_WINDOWEVENT_RESIZED) {
#ifdef DEBUG
                    io.DisplaySize = ImVec2(event.window.data1, event.window.data2);
#endif

                    target = renderer->createSurfaceRenderTarget(surface, target);
                }
            }

            if(event.type == SDL_KEYDOWN && event.key.keysym.scancode == SDL_SCANCODE_F11) {
                if(windowFullscreen == 1) {
                    SDL_SetWindowFullscreen(window, 0);
                    windowFullscreen = 0;
                } else {
                    SDL_SetWindowFullscreen(window, SDL_WINDOW_FULLSCREEN_DESKTOP);
                    windowFullscreen = 1;
                }

                target = renderer->createSurfaceRenderTarget(surface, target);
            }

            if(event.type == SDL_KEYDOWN && event.key.keysym.scancode == SDL_SCANCODE_F7) {
                renderer->takeScreenshot("screenshot.ppm", target);
            }
        }

        if(cinematicMode) {
            animationSystem->update(cinematic, deltaTime);
        } else {
            const Uint8* keys = SDL_GetKeyboardState(nullptr);

            glm::vec3 direction = glm::normalize(cameraTransform->position - cameraComponent->target);
            direction.y = 0;

            glm::vec3 right = glm::normalize(glm::cross(direction, glm::vec3(0, 1, 0)));

            const float playerSpeed = 3;

            if(keys[SDL_SCANCODE_W])
                playerTransform->position -= playerSpeed * direction * deltaTime;

            if(keys[SDL_SCANCODE_S])
                playerTransform->position += playerSpeed * direction * deltaTime;

            if(keys[SDL_SCANCODE_A])
                playerTransform->position -= playerSpeed * right * glm::vec3(1, 0, 0) * deltaTime;

            if(keys[SDL_SCANCODE_D])
                playerTransform->position += playerSpeed * right * glm::vec3(1, 0, 0) * deltaTime;

            cameraComponent->target = playerTransform->position;
        }

        audio->update();

#ifdef DEBUG
        ImGui::NewFrame();

        io.DeltaTime = deltaTime;

        int mouseX = 0, mouseY = 0;
        const Uint32 mouseButtons = SDL_GetMouseState(&mouseX, &mouseY);
        io.MouseDown[0] = (mouseButtons & SDL_BUTTON(SDL_BUTTON_LEFT)) != 0;
        io.MouseDown[1] = (mouseButtons & SDL_BUTTON(SDL_BUTTON_RIGHT)) != 0;
        io.MouseDown[2] = (mouseButtons & SDL_BUTTON(SDL_BUTTON_MIDDLE)) != 0;

        if(SDL_GetWindowFlags(window) & SDL_WINDOW_INPUT_FOCUS)
            io.MousePos = ImVec2(static_cast<float>(mouseX), static_cast<float>(mouseY));

        ImGui::DragFloat("Aperture", &cameraComponent->aperture, 0.1f);
        ImGui::DragFloat("Focus Distance", &cameraComponent->focusDistance, 0.1f);
        ImGui::Render();
#endif

        renderer->render(*worldManager.getCurrentWorld(), target);

        if(record) {
            static int frameNum = 0;
            std::string screenshotName = "frame" + std::to_string(frameNum) + ".ppm";
            frameNum++;

            renderer->takeScreenshot(screenshotName.c_str(), target);
        }
    }

    worldManager.destructResources(renderer);

    renderer->destroyRenderTarget(target);

    vkDestroySurfaceKHR(renderer->getInstance(), surface, nullptr);

    delete renderer;

    ECS::destructResources();

    writeConfig();

    SDL_DestroyWindow(window);

    return 0;
}
