#include "renderer.h"

#include <set>
#include <vector>
#include <iostream>
#include <fstream>
#include <cstring>
#include <cmath>
#include <array>
#include <stb_image.h>

#include "platform.h"
#include "mesh.h"
#include "material.h"
#include "rendercollection.h"
#include "ecs.h"
#include "world.h"

VkResult name_object(VkDevice device, VkObjectType type, uint64_t object,
                     std::string_view name) {
    if (object == 0x0) {
        return VK_ERROR_DEVICE_LOST;
    }

    VkDebugUtilsObjectNameInfoEXT info = {};
    info.sType = VK_STRUCTURE_TYPE_DEBUG_UTILS_OBJECT_NAME_INFO_EXT;
    info.objectType = type;
    info.pObjectName = name.data();
    info.objectHandle = object;

    auto func = (PFN_vkSetDebugUtilsObjectNameEXT) vkGetDeviceProcAddr(
            device, "vkSetDebugUtilsObjectNameEXT");
    if (func != nullptr)
        return func(device, &info);
    else
        return VK_ERROR_EXTENSION_NOT_PRESENT;
}

Renderer::Renderer(GraphicsConfig config) : config_(config) {
    createInstance();
#ifdef DEBUG
    if(enableDebug)
        createDebugMessenger();
#endif
    createLogicalDevice();
    createCommandPool();
    createPresentationRenderPass();
    createDescriptorPool();
    createMaterialSetLayout();
    createEmptyMaterialSet();

    shadowPass_ = new ShadowPass(*this);
    worldPass_ = new WorldPass(*this);
    postPass_ = new PostPass(*this);
    dofPass_ = new DoFPass(*this);
#ifdef DEBUG
    if(config.enableImgui)
        imguiPass_ = new ImGuiPass(*this);
#endif
    skyPass_ = new SkyPass(*this);
    smaaPass_ = new SMAAPass(*this);
    debugPass_ = new DebugPass(*this);
}

Renderer::~Renderer() {
    vkDeviceWaitIdle(device_);

    delete debugPass_;
    delete smaaPass_;
    delete skyPass_;
#ifdef DEBUG
    if(config_.enableImgui)
        delete imguiPass_;
#endif
    delete dofPass_;
    delete postPass_;
    delete worldPass_;
    delete shadowPass_;

    vkDestroySampler(device_, emptySampler_, nullptr);
    vkDestroyImageView(device_, emptyImageView_, nullptr);
    vkFreeMemory(device_, emptyMemory_, nullptr);
    vkDestroyImage(device_, emptyImage_, nullptr);

    vkDestroyDescriptorSetLayout(device_, materialSetLayout_, nullptr);

    vkDestroyDescriptorPool(device_, descriptorPool_, nullptr);

    vkDestroyRenderPass(device_, presentationRenderPass_, nullptr);

    vkDestroyCommandPool(device_, commandPool_, nullptr);

#ifdef DEBUG
    destroyMessenger_(instance_, messenger_, nullptr);
#endif

    vkDestroyDevice(device_, nullptr);

    vkDestroyInstance(instance_, nullptr);
}

void Renderer::render(World& world, RenderTarget* target, RenderExtraInfo* extraInfo) {
    RenderCollection collection;

    // generate collection
    {
        auto meshes = ECS::getWorldComponents<MeshComponent>(&world);
        for(auto [id, mesh] : meshes) {
            RenderMesh renderMesh;
            renderMesh.entity = id;
            renderMesh.mesh = mesh;
            renderMesh.transform = ECS::getComponent<TransformComponent>(id);

            collection.meshes.push_back(renderMesh);
        }

        auto lights = ECS::getWorldComponents<LightComponent>(&world);
        for(auto [id, light] : lights) {
            RenderLight renderLight;
            renderLight.light = light;
            renderLight.transform = ECS::getComponent<TransformComponent>(id);

            collection.lights.push_back(renderLight);
        }

        auto cameras = ECS::getWorldComponents<CameraComponent>(&world);

        auto [id, camera] = cameras[0];

        collection.camera.camera = camera;
        collection.camera.transform = ECS::getComponent<TransformComponent>(id);
    }

    vkAcquireNextImageKHR(device_, target->swapchain, UINT64_MAX, target->imageAvailableSemaphore, nullptr, &target->imageIndex);

    target->currentResource = (target->imageIndex + 1) % numFrameResources;

    vkWaitForFences(device_, 1, &target->fences[target->imageIndex], VK_TRUE, UINT64_MAX);
    vkResetFences(device_, 1, &target->fences[target->imageIndex]);

    VkCommandBuffer commandBuffer = target->commandBuffers[target->currentResource];

    VkCommandBufferBeginInfo beginInfo = {};
    beginInfo.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO;

    vkBeginCommandBuffer(commandBuffer, &beginInfo);

    shadowPass_->render(commandBuffer, collection);

    VkViewport viewport = {};
    viewport.width = target->extent.width;
    viewport.height = target->extent.height;
    viewport.maxDepth = 1.0f;

    vkCmdSetViewport(commandBuffer, 0, 1, &viewport);

    VkRect2D scissor = {};
    scissor.extent = target->extent;

    vkCmdSetScissor(commandBuffer, 0, 1, &scissor);

    std::array<VkClearValue, 2> clearColor = {};
    clearColor[1].depthStencil.depth = 1.0f;

    VkRenderPassBeginInfo renderPassBeginInfo = {};
    renderPassBeginInfo.sType = VK_STRUCTURE_TYPE_RENDER_PASS_BEGIN_INFO;
    renderPassBeginInfo.framebuffer = target->offscreenFramebuffers[target->currentResource];
    renderPassBeginInfo.renderPass = worldPass_->getRenderPass();
    renderPassBeginInfo.renderArea.extent = target->extent;
    renderPassBeginInfo.clearValueCount = clearColor.size();
    renderPassBeginInfo.pClearValues = clearColor.data();

    vkCmdBeginRenderPass(commandBuffer, &renderPassBeginInfo, VK_SUBPASS_CONTENTS_INLINE);

    worldPass_->render(commandBuffer, collection, target);
    skyPass_->render(world.color, commandBuffer);

    vkCmdEndRenderPass(commandBuffer);

    debugPass_->render(commandBuffer, collection, extraInfo, target);

    dofPass_->render(commandBuffer, *collection.camera.camera, target);

    // reset after dof pass
    vkCmdSetViewport(commandBuffer, 0, 1, &viewport);
    vkCmdSetScissor(commandBuffer, 0, 1, &scissor);

    smaaPass_->render(commandBuffer, target);

    clearColor = {};

    renderPassBeginInfo = {};
    renderPassBeginInfo.sType = VK_STRUCTURE_TYPE_RENDER_PASS_BEGIN_INFO;
    renderPassBeginInfo.framebuffer = target->swapchainFramebuffers[target->imageIndex];
    renderPassBeginInfo.renderPass = presentationRenderPass_;
    renderPassBeginInfo.renderArea.extent = target->extent;
    renderPassBeginInfo.clearValueCount = 1;
    renderPassBeginInfo.pClearValues = &clearColor[0];

    vkCmdBeginRenderPass(commandBuffer, &renderPassBeginInfo, VK_SUBPASS_CONTENTS_INLINE);

    postPass_->render(commandBuffer, target);
#ifdef DEBUG
    if(config_.enableImgui)
        imguiPass_->render(commandBuffer, target);
#endif

    vkCmdEndRenderPass(commandBuffer);

    vkEndCommandBuffer(commandBuffer);

    const VkPipelineStageFlags waitStage = VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT;

    VkSubmitInfo submitInfo = {};
    submitInfo.sType = VK_STRUCTURE_TYPE_SUBMIT_INFO;
    submitInfo.waitSemaphoreCount = 1;
    submitInfo.pWaitSemaphores = &target->imageAvailableSemaphore;
    submitInfo.pWaitDstStageMask = &waitStage;
    submitInfo.commandBufferCount = 1;
    submitInfo.pCommandBuffers = &commandBuffer;
    submitInfo.signalSemaphoreCount = 1;
    submitInfo.pSignalSemaphores = &target->renderFinishedSemaphore;

    vkQueueSubmit(graphicsQueue_, 1, &submitInfo, target->fences[target->imageIndex]);

    VkPresentInfoKHR presentInfo = {};
    presentInfo.sType = VK_STRUCTURE_TYPE_PRESENT_INFO_KHR;
    presentInfo.waitSemaphoreCount = 1;
    presentInfo.pWaitSemaphores = &target->renderFinishedSemaphore;
    presentInfo.swapchainCount = 1;
    presentInfo.pSwapchains = &target->swapchain;
    presentInfo.pImageIndices = &target->imageIndex;

    vkQueuePresentKHR(graphicsQueue_, &presentInfo);
}

RenderTarget* Renderer::createSurfaceRenderTarget(VkSurfaceKHR surface, RenderTarget* oldTarget) {
    vkDeviceWaitIdle(device_);

    auto target = new RenderTarget();
    target->surface = surface;

    VkSurfaceCapabilitiesKHR surfaceCapabilities = {};
    vkGetPhysicalDeviceSurfaceCapabilitiesKHR(physicalDevice_, surface, &surfaceCapabilities);

    target->extent = surfaceCapabilities.currentExtent;

    uint32_t surfaceFormatCount = 0;
    vkGetPhysicalDeviceSurfaceFormatsKHR(physicalDevice_, surface, &surfaceFormatCount, nullptr);

    auto surfaceFormats = new VkSurfaceFormatKHR[surfaceFormatCount];
    vkGetPhysicalDeviceSurfaceFormatsKHR(physicalDevice_, surface, &surfaceFormatCount, surfaceFormats);

    uint32_t chosenFormat = 0;
    for(uint32_t i = 0; i < surfaceFormatCount; i++) {
        if(surfaceFormats[i].format == VK_FORMAT_B8G8R8A8_SRGB)
            chosenFormat = i;
    }

    VkBool32 supported = VK_FALSE;
    vkGetPhysicalDeviceSurfaceSupportKHR(physicalDevice_, queueIndices.presentation, surface, &supported);

    VkSwapchainCreateInfoKHR swapchainCreateInfo = {};
    swapchainCreateInfo.sType = VK_STRUCTURE_TYPE_SWAPCHAIN_CREATE_INFO_KHR;
    swapchainCreateInfo.surface = surface;
    swapchainCreateInfo.minImageCount = surfaceCapabilities.minImageCount;
    swapchainCreateInfo.imageColorSpace = surfaceFormats[chosenFormat].colorSpace;
    swapchainCreateInfo.imageFormat = surfaceFormats[chosenFormat].format;
    swapchainCreateInfo.imageExtent = surfaceCapabilities.currentExtent;
    swapchainCreateInfo.imageArrayLayers = 1;
    swapchainCreateInfo.imageUsage = VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT | VK_IMAGE_USAGE_TRANSFER_SRC_BIT;
    swapchainCreateInfo.queueFamilyIndexCount = 1;
    swapchainCreateInfo.pQueueFamilyIndices = &queueIndices.presentation;
    swapchainCreateInfo.preTransform = surfaceCapabilities.currentTransform;
    swapchainCreateInfo.compositeAlpha = VK_COMPOSITE_ALPHA_OPAQUE_BIT_KHR;
    swapchainCreateInfo.presentMode = config_.vsync ? VK_PRESENT_MODE_FIFO_KHR : VK_PRESENT_MODE_IMMEDIATE_KHR;
    swapchainCreateInfo.clipped = VK_TRUE;

    if(oldTarget != nullptr)
        swapchainCreateInfo.oldSwapchain = oldTarget->swapchain;

    vkCreateSwapchainKHR(device_, &swapchainCreateInfo, nullptr, &target->swapchain);

    if(oldTarget != nullptr)
        destroyRenderTarget(oldTarget);

    delete[] surfaceFormats;

    uint32_t swapchainImageCount = 0;
    vkGetSwapchainImagesKHR(device_, target->swapchain, &swapchainImageCount, nullptr);

    target->numImages = swapchainImageCount;

    target->swapchainImages = new VkImage[swapchainImageCount];
    vkGetSwapchainImagesKHR(device_, target->swapchain, &swapchainImageCount, target->swapchainImages);

    // create frame resources
    target->swapchainImageViews = new VkImageView[swapchainImageCount];
    target->swapchainFramebuffers = new VkFramebuffer[swapchainImageCount];
    for(uint32_t i = 0; i < swapchainImageCount; i++) {
        // swapchain image view
        {
            VkImageViewCreateInfo createInfo = {};
            createInfo.sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
            createInfo.image = target->swapchainImages[i];
            createInfo.viewType = VK_IMAGE_VIEW_TYPE_2D;
            createInfo.format = VK_FORMAT_B8G8R8A8_SRGB;
            createInfo.subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
            createInfo.subresourceRange.levelCount = 1;
            createInfo.subresourceRange.layerCount = 1;

            vkCreateImageView(device_, &createInfo, nullptr, &target->swapchainImageViews[i]);
        }

        // swapchain framebuffer
        {
            VkFramebufferCreateInfo framebufferInfo = {};
            framebufferInfo.sType = VK_STRUCTURE_TYPE_FRAMEBUFFER_CREATE_INFO;
            framebufferInfo.renderPass = presentationRenderPass_;
            framebufferInfo.attachmentCount = 1;
            framebufferInfo.pAttachments = &target->swapchainImageViews[i];
            framebufferInfo.width = target->extent.width;
            framebufferInfo.height = target->extent.height;
            framebufferInfo.layers = 1;

            vkCreateFramebuffer(device_, &framebufferInfo, nullptr, &target->swapchainFramebuffers[i]);
        }
    }

    // offscreen
    target->offscreenColorImages = new VkImage[numFrameResources];
    target->offscreenColorMemory = new VkDeviceMemory[numFrameResources];
    target->offscreenColorImageViews = new VkImageView[numFrameResources];

    target->offscreenDepthImages = new VkImage[numFrameResources];
    target->offscreenDepthMemory = new VkDeviceMemory[numFrameResources];
    target->offscreenDepthImageViews = new VkImageView[numFrameResources];

    target->offscreenFramebuffers = new VkFramebuffer[numFrameResources];

    // sobel
    target->sobelImages = new VkImage[numFrameResources];
    target->sobelMemorys = new VkDeviceMemory[numFrameResources];
    target->sobelImageViews = new VkImageView[numFrameResources];

    target->sobelFramebuffers = new VkFramebuffer[numFrameResources];

    // dof
    target->nearFieldImages = new VkImage[numFrameResources];
    target->nearFieldMemory = new VkDeviceMemory[numFrameResources];
    target->nearFieldImageViews = new VkImageView[numFrameResources];
    target->nearFieldFramebuffers = new VkFramebuffer[numFrameResources];

    target->farFieldImages = new VkImage[numFrameResources];
    target->farFieldMemory = new VkDeviceMemory[numFrameResources];
    target->farFieldImageViews = new VkImageView[numFrameResources];
    target->farFieldFramebuffers = new VkFramebuffer[numFrameResources];

    // smaa
    target->edgeImages = new VkImage[numFrameResources];
    target->edgeMemorys = new VkDeviceMemory[numFrameResources];
    target->edgeImageViews = new VkImageView[numFrameResources];
    target->edgeFramebuffers = new VkFramebuffer[numFrameResources];
    target->edgeDescriptorSets = new VkDescriptorSet[numFrameResources];

    target->blendImages = new VkImage[numFrameResources];
    target->blendMemorys = new VkDeviceMemory[numFrameResources];
    target->blendImageViews = new VkImageView[numFrameResources];
    target->blendFramebuffers = new VkFramebuffer[numFrameResources];
    target->blendDescriptorSets = new VkDescriptorSet[numFrameResources];

    // imgui
    target->imguiVertexBuffers = new VkBuffer[numFrameResources];
    target->imguiVertexMemorys = new VkDeviceMemory[numFrameResources];
    target->imguiVertexBufferSizes = new size_t[numFrameResources];
    target->imguiIndexBuffers = new VkBuffer[numFrameResources];
    target->imguiIndexMemorys = new VkDeviceMemory[numFrameResources];
    target->imguiIndexBufferSizes = new size_t[numFrameResources];

    for(uint32_t i = 0; i < numFrameResources; i++) {
        // offscreen image
        {
            VkImageCreateInfo imageCreateInfo = {};
            imageCreateInfo.sType = VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO;
            imageCreateInfo.imageType = VK_IMAGE_TYPE_2D;
            imageCreateInfo.format = VK_FORMAT_R32G32B32A32_SFLOAT;
            imageCreateInfo.extent.width = target->extent.width;
            imageCreateInfo.extent.height = target->extent.height;
            imageCreateInfo.extent.depth = 1;
            imageCreateInfo.mipLevels = 1;
            imageCreateInfo.arrayLayers = 1;
            imageCreateInfo.samples = VK_SAMPLE_COUNT_1_BIT;
            imageCreateInfo.tiling = VK_IMAGE_TILING_OPTIMAL;
            imageCreateInfo.usage = VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT | VK_IMAGE_USAGE_SAMPLED_BIT;

            vkCreateImage(device_, &imageCreateInfo, nullptr, &target->offscreenColorImages[i]);

            name_object(device_, VK_OBJECT_TYPE_IMAGE, (uint64_t)target->offscreenColorImages[i], "offscreen image view");

            VkMemoryRequirements memoryRequirements = {};
            vkGetImageMemoryRequirements(device_, target->offscreenColorImages[i], &memoryRequirements);

            VkMemoryAllocateInfo allocateInfo = {};
            allocateInfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
            allocateInfo.allocationSize = memoryRequirements.size;
            allocateInfo.memoryTypeIndex = findMemoryType(memoryRequirements.memoryTypeBits, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT);

            vkAllocateMemory(device_, &allocateInfo, nullptr, &target->offscreenColorMemory[i]);
            vkBindImageMemory(device_, target->offscreenColorImages[i], target->offscreenColorMemory[i], 0);
        }

        // offscreen image view
        {
            VkImageViewCreateInfo createInfo = {};
            createInfo.sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
            createInfo.image = target->offscreenColorImages[i];
            createInfo.viewType = VK_IMAGE_VIEW_TYPE_2D;
            createInfo.format = VK_FORMAT_R32G32B32A32_SFLOAT;
            createInfo.subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
            createInfo.subresourceRange.levelCount = 1;
            createInfo.subresourceRange.layerCount = 1;

            vkCreateImageView(device_, &createInfo, nullptr, &target->offscreenColorImageViews[i]);
        }

        // offscreen depth image
        {
            VkImageCreateInfo imageCreateInfo = {};
            imageCreateInfo.sType = VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO;
            imageCreateInfo.imageType = VK_IMAGE_TYPE_2D;
            imageCreateInfo.format = VK_FORMAT_D32_SFLOAT;
            imageCreateInfo.extent.width = target->extent.width;
            imageCreateInfo.extent.height = target->extent.height;
            imageCreateInfo.extent.depth = 1;
            imageCreateInfo.mipLevels = 1;
            imageCreateInfo.arrayLayers = 1;
            imageCreateInfo.samples = VK_SAMPLE_COUNT_1_BIT;
            imageCreateInfo.tiling = VK_IMAGE_TILING_OPTIMAL;
            imageCreateInfo.usage = VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT | VK_IMAGE_USAGE_SAMPLED_BIT;

            vkCreateImage(device_, &imageCreateInfo, nullptr, &target->offscreenDepthImages[i]);

            name_object(device_, VK_OBJECT_TYPE_IMAGE, (uint64_t)target->offscreenDepthImages[i], "offscreen depth view");

            VkMemoryRequirements memoryRequirements = {};
            vkGetImageMemoryRequirements(device_, target->offscreenDepthImages[i], &memoryRequirements);

            VkMemoryAllocateInfo allocateInfo = {};
            allocateInfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
            allocateInfo.allocationSize = memoryRequirements.size;
            allocateInfo.memoryTypeIndex = findMemoryType(memoryRequirements.memoryTypeBits, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT);

            vkAllocateMemory(device_, &allocateInfo, nullptr, &target->offscreenDepthMemory[i]);
            vkBindImageMemory(device_, target->offscreenDepthImages[i], target->offscreenDepthMemory[i], 0);
        }

        // offscreen depth image view
        {
            VkImageViewCreateInfo createInfo = {};
            createInfo.sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
            createInfo.image = target->offscreenDepthImages[i];
            createInfo.viewType = VK_IMAGE_VIEW_TYPE_2D;
            createInfo.format = VK_FORMAT_D32_SFLOAT;
            createInfo.subresourceRange.aspectMask = VK_IMAGE_ASPECT_DEPTH_BIT;
            createInfo.subresourceRange.levelCount = 1;
            createInfo.subresourceRange.layerCount = 1;

            vkCreateImageView(device_, &createInfo, nullptr, &target->offscreenDepthImageViews[i]);
        }

        // offscreen framebuffer
        {
            const std::array<VkImageView, 2> attachments = {
                target->offscreenColorImageViews[i],
                target->offscreenDepthImageViews[i]
            };

            VkFramebufferCreateInfo framebufferInfo = {};
            framebufferInfo.sType = VK_STRUCTURE_TYPE_FRAMEBUFFER_CREATE_INFO;
            framebufferInfo.renderPass = worldPass_->getRenderPass();
            framebufferInfo.attachmentCount = attachments.size();
            framebufferInfo.pAttachments = attachments.data();
            framebufferInfo.width = target->extent.width;
            framebufferInfo.height = target->extent.height;
            framebufferInfo.layers = 1;

            vkCreateFramebuffer(device_, &framebufferInfo, nullptr, &target->offscreenFramebuffers[i]);
        }

        // sobel image
        {
            VkImageCreateInfo imageCreateInfo = {};
            imageCreateInfo.sType = VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO;
            imageCreateInfo.imageType = VK_IMAGE_TYPE_2D;
            imageCreateInfo.format = VK_FORMAT_R16G16B16A16_SFLOAT;
            imageCreateInfo.extent.width = target->extent.width;
            imageCreateInfo.extent.height = target->extent.height;
            imageCreateInfo.extent.depth = 1;
            imageCreateInfo.mipLevels = 1;
            imageCreateInfo.arrayLayers = 1;
            imageCreateInfo.samples = VK_SAMPLE_COUNT_1_BIT;
            imageCreateInfo.tiling = VK_IMAGE_TILING_OPTIMAL;
            imageCreateInfo.usage = VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT | VK_IMAGE_USAGE_SAMPLED_BIT;

            vkCreateImage(device_, &imageCreateInfo, nullptr, &target->sobelImages[i]);

            name_object(device_, VK_OBJECT_TYPE_IMAGE, (uint64_t)target->sobelImages[i], "sobel image");

            VkMemoryRequirements memoryRequirements = {};
            vkGetImageMemoryRequirements(device_, target->sobelImages[i], &memoryRequirements);

            VkMemoryAllocateInfo allocateInfo = {};
            allocateInfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
            allocateInfo.allocationSize = memoryRequirements.size;
            allocateInfo.memoryTypeIndex = findMemoryType(memoryRequirements.memoryTypeBits, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT);

            vkAllocateMemory(device_, &allocateInfo, nullptr, &target->sobelMemorys[i]);
            vkBindImageMemory(device_, target->sobelImages[i], target->sobelMemorys[i], 0);
        }

        // sobel image view
        {
            VkImageViewCreateInfo createInfo = {};
            createInfo.sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
            createInfo.image = target->sobelImages[i];
            createInfo.viewType = VK_IMAGE_VIEW_TYPE_2D;
            createInfo.format = VK_FORMAT_R16G16B16A16_SFLOAT;
            createInfo.subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
            createInfo.subresourceRange.levelCount = 1;
            createInfo.subresourceRange.layerCount = 1;

            vkCreateImageView(device_, &createInfo, nullptr, &target->sobelImageViews[i]);
        }

        // sobel framebuffer
        {
            VkFramebufferCreateInfo framebufferInfo = {};
            framebufferInfo.sType = VK_STRUCTURE_TYPE_FRAMEBUFFER_CREATE_INFO;
            framebufferInfo.renderPass = debugPass_->getRenderPass();
            framebufferInfo.attachmentCount = 1;
            framebufferInfo.pAttachments = &target->sobelImageViews[i];
            framebufferInfo.width = target->extent.width;
            framebufferInfo.height = target->extent.height;
            framebufferInfo.layers = 1;

            vkCreateFramebuffer(device_, &framebufferInfo, nullptr, &target->sobelFramebuffers[i]);
        }

        // near field color
        {
            VkImageCreateInfo imageCreateInfo = {};
            imageCreateInfo.sType = VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO;
            imageCreateInfo.imageType = VK_IMAGE_TYPE_2D;
            imageCreateInfo.format = VK_FORMAT_R32G32B32A32_SFLOAT;
            imageCreateInfo.extent.width = target->extent.width / config_.dofDownscale;
            imageCreateInfo.extent.height = target->extent.height / config_.dofDownscale;
            imageCreateInfo.extent.depth = 1;
            imageCreateInfo.mipLevels = 1;
            imageCreateInfo.arrayLayers = 1;
            imageCreateInfo.samples = VK_SAMPLE_COUNT_1_BIT;
            imageCreateInfo.tiling = VK_IMAGE_TILING_OPTIMAL;
            imageCreateInfo.usage = VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT | VK_IMAGE_USAGE_SAMPLED_BIT;

            vkCreateImage(device_, &imageCreateInfo, nullptr, &target->nearFieldImages[i]);

            name_object(device_, VK_OBJECT_TYPE_IMAGE, (uint64_t)target->nearFieldImages[i], "near field image");

            VkMemoryRequirements memoryRequirements = {};
            vkGetImageMemoryRequirements(device_, target->nearFieldImages[i], &memoryRequirements);

            VkMemoryAllocateInfo allocateInfo = {};
            allocateInfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
            allocateInfo.allocationSize = memoryRequirements.size;
            allocateInfo.memoryTypeIndex = findMemoryType(memoryRequirements.memoryTypeBits, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT);

            vkAllocateMemory(device_, &allocateInfo, nullptr, &target->nearFieldMemory[i]);
            vkBindImageMemory(device_, target->nearFieldImages[i], target->nearFieldMemory[i], 0);
        }

        // near field image view
        {
            VkImageViewCreateInfo createInfo = {};
            createInfo.sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
            createInfo.image = target->nearFieldImages[i];
            createInfo.viewType = VK_IMAGE_VIEW_TYPE_2D;
            createInfo.format = VK_FORMAT_R32G32B32A32_SFLOAT;
            createInfo.subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
            createInfo.subresourceRange.levelCount = 1;
            createInfo.subresourceRange.layerCount = 1;

            vkCreateImageView(device_, &createInfo, nullptr, &target->nearFieldImageViews[i]);
        }

        // near field framebuffer
        {
            VkFramebufferCreateInfo framebufferInfo = {};
            framebufferInfo.sType = VK_STRUCTURE_TYPE_FRAMEBUFFER_CREATE_INFO;
            framebufferInfo.renderPass = dofPass_->getRenderPass();
            framebufferInfo.attachmentCount = 1;
            framebufferInfo.pAttachments = &target->nearFieldImageViews[i];
            framebufferInfo.width = target->extent.width / config_.dofDownscale;
            framebufferInfo.height = target->extent.height / config_.dofDownscale;
            framebufferInfo.layers = 1;

            vkCreateFramebuffer(device_, &framebufferInfo, nullptr, &target->nearFieldFramebuffers[i]);
        }

        // far field color
        {
            VkImageCreateInfo imageCreateInfo = {};
            imageCreateInfo.sType = VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO;
            imageCreateInfo.imageType = VK_IMAGE_TYPE_2D;
            imageCreateInfo.format = VK_FORMAT_R32G32B32A32_SFLOAT;
            imageCreateInfo.extent.width = target->extent.width / config_.dofDownscale;
            imageCreateInfo.extent.height = target->extent.height / config_.dofDownscale;
            imageCreateInfo.extent.depth = 1;
            imageCreateInfo.mipLevels = 1;
            imageCreateInfo.arrayLayers = 1;
            imageCreateInfo.samples = VK_SAMPLE_COUNT_1_BIT;
            imageCreateInfo.tiling = VK_IMAGE_TILING_OPTIMAL;
            imageCreateInfo.usage = VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT | VK_IMAGE_USAGE_SAMPLED_BIT;

            vkCreateImage(device_, &imageCreateInfo, nullptr, &target->farFieldImages[i]);

            name_object(device_, VK_OBJECT_TYPE_IMAGE, (uint64_t)target->farFieldImages[i], "far field image");

            VkMemoryRequirements memoryRequirements = {};
            vkGetImageMemoryRequirements(device_, target->farFieldImages[i], &memoryRequirements);

            VkMemoryAllocateInfo allocateInfo = {};
            allocateInfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
            allocateInfo.allocationSize = memoryRequirements.size;
            allocateInfo.memoryTypeIndex = findMemoryType(memoryRequirements.memoryTypeBits, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT);

            vkAllocateMemory(device_, &allocateInfo, nullptr, &target->farFieldMemory[i]);
            vkBindImageMemory(device_, target->farFieldImages[i], target->farFieldMemory[i], 0);
        }

        // far field image view
        {
            VkImageViewCreateInfo createInfo = {};
            createInfo.sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
            createInfo.image = target->farFieldImages[i];
            createInfo.viewType = VK_IMAGE_VIEW_TYPE_2D;
            createInfo.format = VK_FORMAT_R32G32B32A32_SFLOAT;
            createInfo.subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
            createInfo.subresourceRange.levelCount = 1;
            createInfo.subresourceRange.layerCount = 1;

            vkCreateImageView(device_, &createInfo, nullptr, &target->farFieldImageViews[i]);
        }

        // far field framebuffer
        {
            VkFramebufferCreateInfo framebufferInfo = {};
            framebufferInfo.sType = VK_STRUCTURE_TYPE_FRAMEBUFFER_CREATE_INFO;
            framebufferInfo.renderPass = dofPass_->getRenderPass();
            framebufferInfo.attachmentCount = 1;
            framebufferInfo.pAttachments = &target->farFieldImageViews[i];
            framebufferInfo.width = target->extent.width / config_.dofDownscale;
            framebufferInfo.height = target->extent.height / config_.dofDownscale;
            framebufferInfo.layers = 1;

            vkCreateFramebuffer(device_, &framebufferInfo, nullptr, &target->farFieldFramebuffers[i]);
        }

        // edge image
        {
            VkImageCreateInfo imageCreateInfo = {};
            imageCreateInfo.sType = VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO;
            imageCreateInfo.imageType = VK_IMAGE_TYPE_2D;
            imageCreateInfo.format = VK_FORMAT_R16G16B16A16_SFLOAT;
            imageCreateInfo.extent.width = target->extent.width;
            imageCreateInfo.extent.height = target->extent.height;
            imageCreateInfo.extent.depth = 1;
            imageCreateInfo.mipLevels = 1;
            imageCreateInfo.arrayLayers = 1;
            imageCreateInfo.samples = VK_SAMPLE_COUNT_1_BIT;
            imageCreateInfo.tiling = VK_IMAGE_TILING_OPTIMAL;
            imageCreateInfo.usage = VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT | VK_IMAGE_USAGE_SAMPLED_BIT;

            vkCreateImage(device_, &imageCreateInfo, nullptr, &target->edgeImages[i]);

            name_object(device_, VK_OBJECT_TYPE_IMAGE, (uint64_t)target->edgeImages[i], "edge image");

            VkMemoryRequirements memoryRequirements = {};
            vkGetImageMemoryRequirements(device_, target->edgeImages[i], &memoryRequirements);

            VkMemoryAllocateInfo allocateInfo = {};
            allocateInfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
            allocateInfo.allocationSize = memoryRequirements.size;
            allocateInfo.memoryTypeIndex = findMemoryType(memoryRequirements.memoryTypeBits, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT);

            vkAllocateMemory(device_, &allocateInfo, nullptr, &target->edgeMemorys[i]);
            vkBindImageMemory(device_, target->edgeImages[i], target->edgeMemorys[i], 0);
        }

        // edge image view
        {
            VkImageViewCreateInfo createInfo = {};
            createInfo.sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
            createInfo.image = target->edgeImages[i];
            createInfo.viewType = VK_IMAGE_VIEW_TYPE_2D;
            createInfo.format = VK_FORMAT_R16G16B16A16_SFLOAT;
            createInfo.subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
            createInfo.subresourceRange.levelCount = 1;
            createInfo.subresourceRange.layerCount = 1;

            vkCreateImageView(device_, &createInfo, nullptr, &target->edgeImageViews[i]);
        }

        // edge framebuffer
        {
            VkFramebufferCreateInfo framebufferInfo = {};
            framebufferInfo.sType = VK_STRUCTURE_TYPE_FRAMEBUFFER_CREATE_INFO;
            framebufferInfo.renderPass = smaaPass_->getEdgeRenderPass();
            framebufferInfo.attachmentCount = 1;
            framebufferInfo.pAttachments = &target->edgeImageViews[i];
            framebufferInfo.width = target->extent.width;
            framebufferInfo.height = target->extent.height;
            framebufferInfo.layers = 1;

            vkCreateFramebuffer(device_, &framebufferInfo, nullptr, &target->edgeFramebuffers[i]);
        }

        // blend image
        {
            VkImageCreateInfo imageCreateInfo = {};
            imageCreateInfo.sType = VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO;
            imageCreateInfo.imageType = VK_IMAGE_TYPE_2D;
            imageCreateInfo.format = VK_FORMAT_R16G16B16A16_SFLOAT;
            imageCreateInfo.extent.width = target->extent.width;
            imageCreateInfo.extent.height = target->extent.height;
            imageCreateInfo.extent.depth = 1;
            imageCreateInfo.mipLevels = 1;
            imageCreateInfo.arrayLayers = 1;
            imageCreateInfo.samples = VK_SAMPLE_COUNT_1_BIT;
            imageCreateInfo.tiling = VK_IMAGE_TILING_OPTIMAL;
            imageCreateInfo.usage = VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT | VK_IMAGE_USAGE_SAMPLED_BIT;

            vkCreateImage(device_, &imageCreateInfo, nullptr, &target->blendImages[i]);

            name_object(device_, VK_OBJECT_TYPE_IMAGE, (uint64_t)target->blendImages[i], "blend image");

            VkMemoryRequirements memoryRequirements = {};
            vkGetImageMemoryRequirements(device_, target->blendImages[i], &memoryRequirements);

            VkMemoryAllocateInfo allocateInfo = {};
            allocateInfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
            allocateInfo.allocationSize = memoryRequirements.size;
            allocateInfo.memoryTypeIndex = findMemoryType(memoryRequirements.memoryTypeBits, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT);

            vkAllocateMemory(device_, &allocateInfo, nullptr, &target->blendMemorys[i]);
            vkBindImageMemory(device_, target->blendImages[i], target->blendMemorys[i], 0);
        }

        // blend image view
        {
            VkImageViewCreateInfo createInfo = {};
            createInfo.sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
            createInfo.image = target->blendImages[i];
            createInfo.viewType = VK_IMAGE_VIEW_TYPE_2D;
            createInfo.format = VK_FORMAT_R16G16B16A16_SFLOAT;
            createInfo.subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
            createInfo.subresourceRange.levelCount = 1;
            createInfo.subresourceRange.layerCount = 1;

            vkCreateImageView(device_, &createInfo, nullptr, &target->blendImageViews[i]);
        }

        // blend framebuffer
        {
            VkFramebufferCreateInfo framebufferInfo = {};
            framebufferInfo.sType = VK_STRUCTURE_TYPE_FRAMEBUFFER_CREATE_INFO;
            framebufferInfo.renderPass = smaaPass_->getBlendRenderPass();
            framebufferInfo.attachmentCount = 1;
            framebufferInfo.pAttachments = &target->blendImageViews[i];
            framebufferInfo.width = target->extent.width;
            framebufferInfo.height = target->extent.height;
            framebufferInfo.layers = 1;

            vkCreateFramebuffer(device_, &framebufferInfo, nullptr, &target->blendFramebuffers[i]);
        }

        // imgui
        {
            target->imguiVertexBuffers[i] = nullptr;
            target->imguiVertexMemorys[i] = nullptr;
            target->imguiVertexBufferSizes[i] = 0;

            target->imguiIndexBuffers[i] = nullptr;
            target->imguiIndexMemorys[i] = nullptr;
            target->imguiIndexBufferSizes[i] = 0;
        }
    }

    postPass_->createDescriptorSet(target);
    dofPass_->createDescriptorSet(target);
    smaaPass_->createDescriptorSets(target);

    VkCommandBufferAllocateInfo allocateInfo = {};
    allocateInfo.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO;
    allocateInfo.level = VK_COMMAND_BUFFER_LEVEL_PRIMARY;
    allocateInfo.commandPool = commandPool_;
    allocateInfo.commandBufferCount = numFrameResources;

    target->commandBuffers = new VkCommandBuffer[numFrameResources];

    vkAllocateCommandBuffers(device_, &allocateInfo, target->commandBuffers);

    VkSemaphoreCreateInfo semaphoreCreateInfo = {};
    semaphoreCreateInfo.sType = VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO;

    vkCreateSemaphore(device_, &semaphoreCreateInfo, nullptr, &target->imageAvailableSemaphore);
    vkCreateSemaphore(device_, &semaphoreCreateInfo, nullptr, &target->renderFinishedSemaphore);

    VkFenceCreateInfo fenceCreateInfo = {};
    fenceCreateInfo.sType = VK_STRUCTURE_TYPE_FENCE_CREATE_INFO;
    fenceCreateInfo.flags = VK_FENCE_CREATE_SIGNALED_BIT;

    target->fences = new VkFence[swapchainImageCount];

    for(uint32_t i = 0; i < swapchainImageCount; i++)
        vkCreateFence(device_, &fenceCreateInfo, nullptr, &target->fences[i]);

    return target;
}

void Renderer::destroyRenderTarget(RenderTarget* target) {
    vkDeviceWaitIdle(device_);

    for(uint32_t i = 0; i < target->numImages; i++)
        vkDestroyFence(device_, target->fences[i], nullptr);

    delete[] target->fences;

    vkDestroySemaphore(device_, target->renderFinishedSemaphore, nullptr);
    vkDestroySemaphore(device_, target->imageAvailableSemaphore, nullptr);

    vkFreeCommandBuffers(device_, commandPool_, target->numImages, target->commandBuffers);

    delete[] target->commandBuffers;

    vkFreeDescriptorSets(device_, descriptorPool_, target->numImages, target->blendDescriptorSets);
    vkFreeDescriptorSets(device_, descriptorPool_, target->numImages, target->edgeDescriptorSets);
    vkFreeDescriptorSets(device_, descriptorPool_, target->numImages, target->dofSets);
    vkFreeDescriptorSets(device_, descriptorPool_, target->numImages, target->postSets);

    for(uint32_t i = 0; i < numFrameResources; i++) {
        // imgui
        vkFreeMemory(device_, target->imguiIndexMemorys[i], nullptr);
        vkDestroyBuffer(device_, target->imguiIndexBuffers[i], nullptr);

        vkFreeMemory(device_, target->imguiVertexMemorys[i], nullptr);
        vkDestroyBuffer(device_, target->imguiVertexBuffers[i], nullptr);

        // smaa
        vkDestroyFramebuffer(device_, target->blendFramebuffers[i], nullptr);

        vkDestroyImageView(device_, target->blendImageViews[i], nullptr);
        vkFreeMemory(device_, target->blendMemorys[i], nullptr);
        vkDestroyImage(device_, target->blendImages[i], nullptr);

        vkDestroyFramebuffer(device_, target->edgeFramebuffers[i], nullptr);

        vkDestroyImageView(device_, target->edgeImageViews[i], nullptr);
        vkFreeMemory(device_, target->edgeMemorys[i], nullptr);
        vkDestroyImage(device_, target->edgeImages[i], nullptr);

        // dof
        vkDestroyFramebuffer(device_, target->nearFieldFramebuffers[i], nullptr);

        vkDestroyImageView(device_, target->nearFieldImageViews[i], nullptr);
        vkFreeMemory(device_, target->nearFieldMemory[i], nullptr);
        vkDestroyImage(device_, target->nearFieldImages[i], nullptr);

        vkDestroyFramebuffer(device_, target->farFieldFramebuffers[i], nullptr);

        vkDestroyImageView(device_, target->farFieldImageViews[i], nullptr);
        vkFreeMemory(device_, target->farFieldMemory[i], nullptr);
        vkDestroyImage(device_, target->farFieldImages[i], nullptr);

        // sobel
        vkDestroyFramebuffer(device_, target->sobelFramebuffers[i], nullptr);

        vkDestroyImageView(device_, target->sobelImageViews[i], nullptr);
        vkFreeMemory(device_, target->sobelMemorys[i], nullptr);
        vkDestroyImage(device_, target->sobelImages[i], nullptr);

        // offscreen
        vkDestroyFramebuffer(device_, target->offscreenFramebuffers[i], nullptr);

        vkDestroyImageView(device_, target->offscreenDepthImageViews[i], nullptr);
        vkFreeMemory(device_, target->offscreenDepthMemory[i], nullptr);
        vkDestroyImage(device_, target->offscreenDepthImages[i], nullptr);

        vkDestroyImageView(device_, target->offscreenColorImageViews[i], nullptr);
        vkFreeMemory(device_, target->offscreenColorMemory[i], nullptr);
        vkDestroyImage(device_, target->offscreenColorImages[i], nullptr);
    }

    for(uint32_t i =0; i < target->numImages; i++) {
        vkDestroyFramebuffer(device_, target->swapchainFramebuffers[i], nullptr);
        vkDestroyImageView(device_, target->swapchainImageViews[i], nullptr);
    }

    // imgui
    delete[] target->imguiIndexBufferSizes;
    delete[] target->imguiIndexMemorys;
    delete[] target->imguiIndexBuffers;

    delete[] target->imguiVertexBufferSizes;
    delete[] target->imguiVertexMemorys;
    delete[] target->imguiVertexBuffers;

    // smaa
    delete[] target->blendDescriptorSets;

    delete[] target->blendFramebuffers;
    delete[] target->blendImageViews;
    delete[] target->blendMemorys;
    delete[] target->blendImages;

    delete[] target->edgeDescriptorSets;

    delete[] target->edgeFramebuffers;
    delete[] target->edgeImageViews;
    delete[] target->edgeMemorys;
    delete[] target->edgeImages;

    // dof
    delete[] target->nearFieldFramebuffers;
    delete[] target->nearFieldImageViews;
    delete[] target->nearFieldMemory;
    delete[] target->nearFieldImages;

    delete[] target->farFieldFramebuffers;
    delete[] target->farFieldImageViews;
    delete[] target->farFieldMemory;
    delete[] target->farFieldImages;

    // sobel
    delete[] target->sobelFramebuffers;
    delete[] target->sobelImageViews;
    delete[] target->sobelMemorys;
    delete[] target->sobelImages;

    // offscreen
    delete[] target->offscreenFramebuffers;

    delete[] target->offscreenDepthImageViews;
    delete[] target->offscreenDepthMemory;
    delete[] target->offscreenDepthImages;

    delete[] target->offscreenColorImageViews;
    delete[] target->offscreenColorMemory;
    delete[] target->offscreenColorImages;

    // swapchain
    delete[] target->swapchainFramebuffers;
    delete[] target->swapchainImageViews;
    delete[] target->swapchainImages;

    delete[] target->dofSets;
    delete[] target->postSets;

    vkDestroySwapchainKHR(device_, target->swapchain, nullptr);

    delete target;
}

void Renderer::takeScreenshot(const char* path, RenderTarget* target) {
    VkImageCreateInfo imageCreateInfo = {};
    imageCreateInfo.sType = VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO;
    imageCreateInfo.imageType = VK_IMAGE_TYPE_2D;
    imageCreateInfo.format = VK_FORMAT_R8G8B8A8_UNORM;
    imageCreateInfo.extent.width = target->extent.width;
    imageCreateInfo.extent.height = target->extent.height;
    imageCreateInfo.extent.depth = 1;
    imageCreateInfo.mipLevels = 1;
    imageCreateInfo.arrayLayers = 1;
    imageCreateInfo.samples = VK_SAMPLE_COUNT_1_BIT;
    imageCreateInfo.tiling = VK_IMAGE_TILING_LINEAR;
    imageCreateInfo.usage = VK_IMAGE_USAGE_TRANSFER_DST_BIT;

    VkImage image = nullptr;
    vkCreateImage(device_, &imageCreateInfo, nullptr, &image);

    name_object(device_, VK_OBJECT_TYPE_IMAGE, (uint64_t)image, "screenshot image");

    VkMemoryRequirements memoryRequirements = {};
    vkGetImageMemoryRequirements(device_, image, &memoryRequirements);

    VkMemoryAllocateInfo allocateInfo = {};
    allocateInfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
    allocateInfo.allocationSize = memoryRequirements.size;
    allocateInfo.memoryTypeIndex = findMemoryType(memoryRequirements.memoryTypeBits, VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT);

    VkDeviceMemory imageMemory = nullptr;
    vkAllocateMemory(device_, &allocateInfo, nullptr, &imageMemory);
    vkBindImageMemory(device_, image, imageMemory, 0);

    VkCommandBufferAllocateInfo bufferAllocateInfo = {};
    bufferAllocateInfo.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO;
    bufferAllocateInfo.commandPool = commandPool_;
    bufferAllocateInfo.commandBufferCount = 1;

    VkCommandBuffer commandBuffer = nullptr;
    vkAllocateCommandBuffers(device_, &bufferAllocateInfo, &commandBuffer);

    VkCommandBufferBeginInfo beginInfo = {};
    beginInfo.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO;

    vkBeginCommandBuffer(commandBuffer, &beginInfo);

    // change screenshot image to transfer dst layout
    {
        VkImageMemoryBarrier imageMemoryBarrier = {};
        imageMemoryBarrier.sType = VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER;
        imageMemoryBarrier.dstAccessMask = VK_ACCESS_TRANSFER_WRITE_BIT;
        imageMemoryBarrier.newLayout = VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL;
        imageMemoryBarrier.image = image;
        imageMemoryBarrier.subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
        imageMemoryBarrier.subresourceRange.layerCount = 1;
        imageMemoryBarrier.subresourceRange.levelCount = 1;

        vkCmdPipelineBarrier(
            commandBuffer,
            VK_PIPELINE_STAGE_TRANSFER_BIT,
            VK_PIPELINE_STAGE_TRANSFER_BIT,
            0,
            0, nullptr,
            0, nullptr,
            1, &imageMemoryBarrier);
    }

    VkImage srcImage = target->swapchainImages[0]; // FIXME: use previous image

    // change offscreen image to transfer src layout
    {
        VkImageMemoryBarrier imageMemoryBarrier = {};
        imageMemoryBarrier.sType = VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER;
        imageMemoryBarrier.srcAccessMask = VK_ACCESS_MEMORY_READ_BIT;
        imageMemoryBarrier.dstAccessMask = VK_ACCESS_TRANSFER_READ_BIT;
        imageMemoryBarrier.oldLayout = VK_IMAGE_LAYOUT_PRESENT_SRC_KHR;
        imageMemoryBarrier.newLayout = VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL;
        imageMemoryBarrier.image = srcImage;
        imageMemoryBarrier.subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
        imageMemoryBarrier.subresourceRange.layerCount = 1;
        imageMemoryBarrier.subresourceRange.levelCount = 1;

        vkCmdPipelineBarrier(
            commandBuffer,
            VK_PIPELINE_STAGE_TRANSFER_BIT,
            VK_PIPELINE_STAGE_TRANSFER_BIT,
            0,
            0, nullptr,
            0, nullptr,
            1, &imageMemoryBarrier);
    }

    VkImageCopy imageCopy = {};
    imageCopy.srcSubresource.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
    imageCopy.srcSubresource.layerCount = 1;
    imageCopy.dstSubresource.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
    imageCopy.dstSubresource.layerCount = 1;
    imageCopy.extent.width = target->extent.width;
    imageCopy.extent.height = target->extent.height;
    imageCopy.extent.depth = 1;

    // Issue the copy command
    vkCmdCopyImage(
        commandBuffer,
        srcImage,
        VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL,
        image,
        VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
        1,
        &imageCopy);

    // change screenshot image from transfer dst to general (for mapping memory)
    {
        VkImageMemoryBarrier imageMemoryBarrier = {};
        imageMemoryBarrier.sType = VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER;
        imageMemoryBarrier.srcAccessMask = VK_ACCESS_TRANSFER_WRITE_BIT;
        imageMemoryBarrier.dstAccessMask = VK_ACCESS_MEMORY_READ_BIT;
        imageMemoryBarrier.oldLayout = VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL;
        imageMemoryBarrier.newLayout = VK_IMAGE_LAYOUT_GENERAL;
        imageMemoryBarrier.image = image;
        imageMemoryBarrier.subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
        imageMemoryBarrier.subresourceRange.layerCount = 1;
        imageMemoryBarrier.subresourceRange.levelCount = 1;

        vkCmdPipelineBarrier(
            commandBuffer,
            VK_PIPELINE_STAGE_TRANSFER_BIT,
            VK_PIPELINE_STAGE_TRANSFER_BIT,
            0,
            0, nullptr,
            0, nullptr,
            1, &imageMemoryBarrier);
    }

    // change offscreen image layout back to normal
    {
        VkImageMemoryBarrier imageMemoryBarrier = {};
        imageMemoryBarrier.sType = VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER;
        imageMemoryBarrier.srcAccessMask = VK_ACCESS_TRANSFER_READ_BIT;
        imageMemoryBarrier.dstAccessMask = VK_ACCESS_MEMORY_READ_BIT;
        imageMemoryBarrier.oldLayout = VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL;
        imageMemoryBarrier.newLayout = VK_IMAGE_LAYOUT_PRESENT_SRC_KHR;
        imageMemoryBarrier.image = srcImage;
        imageMemoryBarrier.subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
        imageMemoryBarrier.subresourceRange.layerCount = 1;
        imageMemoryBarrier.subresourceRange.levelCount = 1;

        vkCmdPipelineBarrier(
            commandBuffer,
            VK_PIPELINE_STAGE_TRANSFER_BIT,
            VK_PIPELINE_STAGE_TRANSFER_BIT,
            0,
            0, nullptr,
            0, nullptr,
            1, &imageMemoryBarrier);
    }

    vkEndCommandBuffer(commandBuffer);

    VkSubmitInfo submitInfo = {};
    submitInfo.sType = VK_STRUCTURE_TYPE_SUBMIT_INFO;
    submitInfo.commandBufferCount = 1;
    submitInfo.pCommandBuffers = &commandBuffer;

    VkFenceCreateInfo fenceCreateInfo = {};
    fenceCreateInfo.sType = VK_STRUCTURE_TYPE_FENCE_CREATE_INFO;

    VkFence fence = nullptr;
    vkCreateFence(device_, &fenceCreateInfo, nullptr, &fence);

    vkQueueSubmit(graphicsQueue_, 1, &submitInfo, fence);

    vkWaitForFences(device_, 1, &fence, VK_TRUE, -1);
    vkDestroyFence(device_, fence, nullptr);

    vkFreeCommandBuffers(device_, commandPool_, 1, &commandBuffer);

    VkImageSubresource subResource = {};
    subResource.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;

    VkSubresourceLayout subResourceLayout = {};
    vkGetImageSubresourceLayout(device_, image, &subResource, &subResourceLayout);

    const char* data = nullptr;
    vkMapMemory(device_, imageMemory, 0, VK_WHOLE_SIZE, 0, (void**)&data);
    data += subResourceLayout.offset;

    std::ofstream file(path, std::ios::out | std::ios::binary);
    file << "P6\n" << target->extent.width << "\n" << target->extent.height << "\n" << 255 << "\n";

    for(uint32_t y = 0; y < target->extent.height; y++) {
        auto row = reinterpret_cast<const unsigned int*>(data);
        for(uint32_t x = 0; x < target->extent.width; x++) {
            file.write(reinterpret_cast<const char*>(row), 3);
            row++;
        }

        data += subResourceLayout.rowPitch;
    }

    file.close();

    vkUnmapMemory(device_, imageMemory);

    vkFreeMemory(device_, imageMemory, nullptr);
    vkDestroyImage(device_, image, nullptr);
}

VkShaderModule Renderer::createShader(const char* path) {
    std::ifstream file(path, std::ios::ate | std::ios::binary);

    size_t fileSize = (size_t) file.tellg();
    std::vector<char> buffer(fileSize);
    file.seekg(0);
    file.read(buffer.data(), fileSize);
    file.close();

    VkShaderModuleCreateInfo createInfo = {};
    createInfo.sType = VK_STRUCTURE_TYPE_SHADER_MODULE_CREATE_INFO;
    createInfo.codeSize = buffer.size();
    createInfo.pCode = reinterpret_cast<const uint32_t*>(buffer.data());

    VkShaderModule shaderModule;
    vkCreateShaderModule(device_, &createInfo, nullptr, &shaderModule);

    return shaderModule;
}

void Renderer::uploadImageData(VkImage image, int width, int height, VkDeviceSize size, const void* src) {
    VkBuffer stagingBuffer;
    VkDeviceMemory stagingMemory;

    VkBufferCreateInfo bufferInfo = {};
    bufferInfo.sType = VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO;
    bufferInfo.size = size;
    bufferInfo.usage = VK_BUFFER_USAGE_TRANSFER_SRC_BIT;
    bufferInfo.sharingMode = VK_SHARING_MODE_EXCLUSIVE;

    vkCreateBuffer(device_, &bufferInfo, nullptr, &stagingBuffer);

    VkMemoryRequirements memRequirements = {};
    vkGetBufferMemoryRequirements(device_, stagingBuffer, &memRequirements);

    VkMemoryAllocateInfo allocInfo = {};
    allocInfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
    allocInfo.allocationSize = memRequirements.size;
    allocInfo.memoryTypeIndex = findMemoryType(memRequirements.memoryTypeBits, VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT);

    vkAllocateMemory(device_, &allocInfo, nullptr, &stagingMemory);
    vkBindBufferMemory(device_, stagingBuffer, stagingMemory, 0);

    void* data;
    vkMapMemory(device_, stagingMemory, 0, size, 0, &data);
    memcpy(data, src, size);
    vkUnmapMemory(device_, stagingMemory);

    VkCommandBufferAllocateInfo bufferAllocateInfo = {};
    bufferAllocateInfo.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO;
    bufferAllocateInfo.commandPool = commandPool_;
    bufferAllocateInfo.commandBufferCount = 1;

    VkCommandBuffer commandBuffer = nullptr;
    vkAllocateCommandBuffers(device_, &bufferAllocateInfo, &commandBuffer);

    VkCommandBufferBeginInfo beginInfo = {};
    beginInfo.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO;

    vkBeginCommandBuffer(commandBuffer, &beginInfo);

    // change layout to transfer dst
    {
        VkImageMemoryBarrier imageMemoryBarrier = {};
        imageMemoryBarrier.sType = VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER;
        imageMemoryBarrier.dstAccessMask = VK_ACCESS_TRANSFER_WRITE_BIT;
        imageMemoryBarrier.newLayout = VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL;
        imageMemoryBarrier.image = image;
        imageMemoryBarrier.subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
        imageMemoryBarrier.subresourceRange.layerCount = 1;
        imageMemoryBarrier.subresourceRange.levelCount = 1;

        vkCmdPipelineBarrier(
            commandBuffer,
            VK_PIPELINE_STAGE_TRANSFER_BIT,
            VK_PIPELINE_STAGE_TRANSFER_BIT,
            0,
            0, nullptr,
            0, nullptr,
            1, &imageMemoryBarrier);
    }

    VkBufferImageCopy region = {};
    region.imageSubresource.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
    region.imageSubresource.layerCount = 1;
    region.imageExtent.width = width;
    region.imageExtent.height = height;
    region.imageExtent.depth = 1;

    vkCmdCopyBufferToImage(
        commandBuffer,
        stagingBuffer,
        image,
        VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
        1,
        &region
    );

    // change layout to shader read only
    {
        VkImageMemoryBarrier imageMemoryBarrier = {};
        imageMemoryBarrier.sType = VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER;
        imageMemoryBarrier.srcAccessMask = VK_ACCESS_TRANSFER_WRITE_BIT;
        imageMemoryBarrier.dstAccessMask = VK_ACCESS_SHADER_READ_BIT;
        imageMemoryBarrier.oldLayout = VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL;
        imageMemoryBarrier.newLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
        imageMemoryBarrier.image = image;
        imageMemoryBarrier.subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
        imageMemoryBarrier.subresourceRange.layerCount = 1;
        imageMemoryBarrier.subresourceRange.levelCount = 1;

        vkCmdPipelineBarrier(
            commandBuffer,
            VK_PIPELINE_STAGE_TRANSFER_BIT,
            VK_PIPELINE_STAGE_FRAGMENT_SHADER_BIT,
            0,
            0, nullptr,
            0, nullptr,
            1, &imageMemoryBarrier);
    }

    vkEndCommandBuffer(commandBuffer);

    VkSubmitInfo submitInfo = {};
    submitInfo.sType = VK_STRUCTURE_TYPE_SUBMIT_INFO;
    submitInfo.commandBufferCount = 1;
    submitInfo.pCommandBuffers = &commandBuffer;

    VkFenceCreateInfo fenceCreateInfo = {};
    fenceCreateInfo.sType = VK_STRUCTURE_TYPE_FENCE_CREATE_INFO;

    VkFence fence = nullptr;
    vkCreateFence(device_, &fenceCreateInfo, nullptr, &fence);

    vkQueueSubmit(graphicsQueue_, 1, &submitInfo, fence);

    vkWaitForFences(device_, 1, &fence, VK_TRUE, -1);
    vkDestroyFence(device_, fence, nullptr);

    vkFreeCommandBuffers(device_, commandPool_, 1, &commandBuffer);

    vkFreeMemory(device_, stagingMemory, nullptr);
    vkDestroyBuffer(device_, stagingBuffer, nullptr);
}


uint32_t Renderer::findMemoryType(const uint32_t typeFilter, const VkMemoryPropertyFlags properties) {
    for(uint32_t i = 0; i < deviceMemoryProperties_.memoryTypeCount; i++) {
        if((typeFilter & (1 << i)) && (deviceMemoryProperties_.memoryTypes[i].propertyFlags & properties) == properties) {
            return i;
        }
    }

    return 0;
}

void Renderer::fillMeshBuffers(MeshAsset* mesh) {
    // vertex
    {
        VkBufferCreateInfo bufferInfo = {};
        bufferInfo.sType = VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO;
        bufferInfo.size = sizeof(Vertex) * mesh->vertices.size();
        bufferInfo.usage = VK_BUFFER_USAGE_VERTEX_BUFFER_BIT;
        bufferInfo.sharingMode = VK_SHARING_MODE_EXCLUSIVE;

        vkCreateBuffer(device_, &bufferInfo, nullptr, &mesh->vertexBuffer);

        VkMemoryRequirements memRequirements;
        vkGetBufferMemoryRequirements(device_, mesh->vertexBuffer, &memRequirements);

        VkMemoryAllocateInfo allocInfo = {};
        allocInfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
        allocInfo.allocationSize = memRequirements.size;
        allocInfo.memoryTypeIndex = findMemoryType(memRequirements.memoryTypeBits, VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT);

        vkAllocateMemory(device_, &allocInfo, nullptr, &mesh->vertexMemory);
        vkBindBufferMemory(device_, mesh->vertexBuffer, mesh->vertexMemory, 0);

        void* data;
        vkMapMemory(device_, mesh->vertexMemory, 0, bufferInfo.size, 0, &data);
            memcpy(data, mesh->vertices.data(), bufferInfo.size);
        vkUnmapMemory(device_, mesh->vertexMemory);
    }

    // index
    {
        VkBufferCreateInfo bufferInfo = {};
        bufferInfo.sType = VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO;
        bufferInfo.size = sizeof(uint32_t) * mesh->indices.size();
        bufferInfo.usage = VK_BUFFER_USAGE_INDEX_BUFFER_BIT;
        bufferInfo.sharingMode = VK_SHARING_MODE_EXCLUSIVE;

        vkCreateBuffer(device_, &bufferInfo, nullptr, &mesh->indexBuffer);

        VkMemoryRequirements memRequirements;
        vkGetBufferMemoryRequirements(device_, mesh->indexBuffer, &memRequirements);

        VkMemoryAllocateInfo allocInfo = {};
        allocInfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
        allocInfo.allocationSize = memRequirements.size;
        allocInfo.memoryTypeIndex = findMemoryType(memRequirements.memoryTypeBits, VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT);

        vkAllocateMemory(device_, &allocInfo, nullptr, &mesh->indexMemory);
        vkBindBufferMemory(device_, mesh->indexBuffer, mesh->indexMemory, 0);

        void* data;
        vkMapMemory(device_, mesh->indexMemory, 0, bufferInfo.size, 0, &data);
        memcpy(data, mesh->indices.data(), bufferInfo.size);
        vkUnmapMemory(device_, mesh->indexMemory);
    }
}

void Renderer::fillMaterialBuffers(MaterialAsset* material) {
    int width = 0, height = 0, channels = 0;
    stbi_uc* pixels = stbi_load(material->albedoTexturePath.c_str(), &width, &height, &channels, STBI_rgb_alpha);
    if(pixels == nullptr) {
        material->set = emptyMaterialSet_;

        return;
    }

    VkImageCreateInfo imageCreateInfo = {};
    imageCreateInfo.sType = VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO;
    imageCreateInfo.imageType = VK_IMAGE_TYPE_2D;
    imageCreateInfo.extent.width = width;
    imageCreateInfo.extent.height = height;
    imageCreateInfo.extent.depth = 1;
    imageCreateInfo.mipLevels = 1;
    imageCreateInfo.arrayLayers = 1;
    imageCreateInfo.format = VK_FORMAT_R8G8B8A8_UNORM;
    imageCreateInfo.tiling = VK_IMAGE_TILING_OPTIMAL;
    imageCreateInfo.usage = VK_IMAGE_USAGE_TRANSFER_DST_BIT | VK_IMAGE_USAGE_SAMPLED_BIT;
    imageCreateInfo.sharingMode = VK_SHARING_MODE_EXCLUSIVE;
    imageCreateInfo.samples = VK_SAMPLE_COUNT_1_BIT;

    vkCreateImage(device_, &imageCreateInfo, nullptr, &material->albedoImage);

    name_object(device_, VK_OBJECT_TYPE_IMAGE, (uint64_t)material->albedoImage, "albedo image");

    VkMemoryRequirements memRequirements;
    vkGetImageMemoryRequirements(device_, material->albedoImage, &memRequirements);

    VkMemoryAllocateInfo allocInfo = {};
    allocInfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
    allocInfo.allocationSize = memRequirements.size;
    allocInfo.memoryTypeIndex = findMemoryType(memRequirements.memoryTypeBits, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT);

    vkAllocateMemory(device_, &allocInfo, nullptr, &material->albedoMemory);
    vkBindImageMemory(device_, material->albedoImage, material->albedoMemory, 0);

    uploadImageData(
        material->albedoImage,
        width,
        height,
        width * height * 4,
        pixels
    );

    stbi_image_free(pixels);

    VkImageViewCreateInfo createInfo = {};
    createInfo.sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
    createInfo.image = material->albedoImage;
    createInfo.viewType = VK_IMAGE_VIEW_TYPE_2D;
    createInfo.format = VK_FORMAT_R8G8B8A8_UNORM;
    createInfo.subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
    createInfo.subresourceRange.levelCount = 1;
    createInfo.subresourceRange.layerCount = 1;

    vkCreateImageView(device_, &createInfo, nullptr, &material->albedoImageView);

    VkSamplerCreateInfo samplerInfo = {};
    samplerInfo.sType = VK_STRUCTURE_TYPE_SAMPLER_CREATE_INFO;
    samplerInfo.magFilter = VK_FILTER_LINEAR;
    samplerInfo.minFilter = VK_FILTER_LINEAR;
    samplerInfo.addressModeU = VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_BORDER;
    samplerInfo.addressModeV = VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_BORDER;
    samplerInfo.addressModeW = VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_BORDER;
    samplerInfo.borderColor = VK_BORDER_COLOR_INT_OPAQUE_BLACK;
    samplerInfo.mipmapMode = VK_SAMPLER_MIPMAP_MODE_LINEAR;

    vkCreateSampler(device_, &samplerInfo, nullptr, &material->albedoSampler);

    VkDescriptorSetAllocateInfo setAllocInfo = {};
    setAllocInfo.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_ALLOCATE_INFO;
    setAllocInfo.descriptorPool = descriptorPool_;
    setAllocInfo.descriptorSetCount = 1;
    setAllocInfo.pSetLayouts = &materialSetLayout_;

    vkAllocateDescriptorSets(device_, &setAllocInfo, &material->set);

    VkDescriptorImageInfo albedoImageInfo = {};
    albedoImageInfo.imageLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
    albedoImageInfo.imageView = material->albedoImageView;
    albedoImageInfo.sampler = material->albedoSampler;

    VkWriteDescriptorSet albedoDescriptorWrite = {};
    albedoDescriptorWrite.sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
    albedoDescriptorWrite.descriptorCount = 1;
    albedoDescriptorWrite.descriptorType = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
    albedoDescriptorWrite.dstSet = material->set;
    albedoDescriptorWrite.pImageInfo = &albedoImageInfo;

    vkUpdateDescriptorSets(device_, 1, &albedoDescriptorWrite, 0, nullptr);
}

void Renderer::destroyMaterialBuffers(MaterialAsset* material) {
    vkDeviceWaitIdle(device_);

    vkFreeDescriptorSets(device_, descriptorPool_, 1, &material->set);

    vkDestroySampler(device_, material->albedoSampler, nullptr);
    vkDestroyImageView(device_, material->albedoImageView, nullptr);
    vkFreeMemory(device_, material->albedoMemory, nullptr);
    vkDestroyImage(device_, material->albedoImage, nullptr);
}

void Renderer::destroyMeshBuffers(MeshAsset* mesh) {
    vkDeviceWaitIdle(device_);

    vkFreeMemory(device_, mesh->indexMemory, nullptr);
    vkDestroyBuffer(device_, mesh->indexBuffer, nullptr);

    vkFreeMemory(device_, mesh->vertexMemory, nullptr);
    vkDestroyBuffer(device_, mesh->vertexBuffer, nullptr);
}

void Renderer::createInstance() {
    std::vector<const char*> enabledLayers;
#ifdef DEBUG
    uint32_t layerCount = 0;
    vkEnumerateInstanceLayerProperties(&layerCount, nullptr);

    auto availableLayers = new VkLayerProperties[layerCount];
    vkEnumerateInstanceLayerProperties(&layerCount, availableLayers);

    for(uint32_t i = 0; i < layerCount; i++) {
        if(strcmp(availableLayers[i].layerName, "VK_LAYER_LUNARG_standard_validation") == 0)
            enabledLayers.push_back("VK_LAYER_LUNARG_standard_validation");
    }

    delete[] availableLayers;
#endif

    std::vector<const char*> enabledExtensions;
#ifdef DEBUG
    uint32_t extensionCount = 0;
    vkEnumerateInstanceExtensionProperties(nullptr, &extensionCount, nullptr);

    auto availableExtensions = new VkExtensionProperties[extensionCount];
    vkEnumerateInstanceExtensionProperties(nullptr, &extensionCount, availableExtensions);

    for(uint32_t i = 0; i < extensionCount; i++) {
        if(strcmp(availableExtensions[i].extensionName, "VK_EXT_debug_utils") == 0) {
            enabledExtensions.push_back("VK_EXT_debug_utils");
            enableDebug = true;
        }
    }

    delete[] availableExtensions;
#endif

    auto requiredExtensions = platform::getRequiredExtensions();
    enabledExtensions.insert(enabledExtensions.end(), requiredExtensions.begin(), requiredExtensions.end());

    VkInstanceCreateInfo instanceCreateInfo = {};
    instanceCreateInfo.sType = VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO;
    instanceCreateInfo.enabledLayerCount = enabledLayers.size();
    instanceCreateInfo.ppEnabledLayerNames = enabledLayers.data();
    instanceCreateInfo.enabledExtensionCount = enabledExtensions.size();
    instanceCreateInfo.ppEnabledExtensionNames = enabledExtensions.data();

    vkCreateInstance(&instanceCreateInfo, nullptr, &instance_);
}

#ifdef DEBUG
void Renderer::createDebugMessenger() {
    createMessenger_ = (PFN_vkCreateDebugUtilsMessengerEXT)vkGetInstanceProcAddr(instance_, "vkCreateDebugUtilsMessengerEXT");
    destroyMessenger_ = (PFN_vkDestroyDebugUtilsMessengerEXT)vkGetInstanceProcAddr(instance_, "vkDestroyDebugUtilsMessengerEXT");

    VkDebugUtilsMessengerCreateInfoEXT messengerCreateInfo = {};
    messengerCreateInfo.sType = VK_STRUCTURE_TYPE_DEBUG_UTILS_MESSENGER_CREATE_INFO_EXT;
    messengerCreateInfo.messageSeverity = VK_DEBUG_UTILS_MESSAGE_SEVERITY_ERROR_BIT_EXT | VK_DEBUG_UTILS_MESSAGE_SEVERITY_WARNING_BIT_EXT;
    messengerCreateInfo.messageType = VK_DEBUG_UTILS_MESSAGE_TYPE_GENERAL_BIT_EXT | VK_DEBUG_UTILS_MESSAGE_TYPE_VALIDATION_BIT_EXT | VK_DEBUG_UTILS_MESSAGE_TYPE_PERFORMANCE_BIT_EXT;
    messengerCreateInfo.pfnUserCallback = [](VkDebugUtilsMessageSeverityFlagBitsEXT, unsigned int, const VkDebugUtilsMessengerCallbackDataEXT* callback, void*) -> unsigned int {
        std::cout << callback->pMessage << std::endl;

        return VK_SUCCESS;
    };

    createMessenger_(instance_, &messengerCreateInfo, nullptr, &messenger_);
}
#endif

void Renderer::createLogicalDevice() {
    uint32_t physicalDeviceCount = 0;
    vkEnumeratePhysicalDevices(instance_, &physicalDeviceCount, nullptr);

    auto physicalDevices = new VkPhysicalDevice[physicalDeviceCount];
    vkEnumeratePhysicalDevices(instance_, &physicalDeviceCount, physicalDevices);

    for(uint32_t i = 0; i < physicalDeviceCount; i++) {
        VkPhysicalDeviceProperties properties = {};
        vkGetPhysicalDeviceProperties(physicalDevices[i], &properties);

        physicalDevice_ = physicalDevices[i];
    }

    delete[] physicalDevices;

    vkGetPhysicalDeviceMemoryProperties(physicalDevice_, &deviceMemoryProperties_);

    uint32_t queueFamilyPropertiesCount = 0;
    vkGetPhysicalDeviceQueueFamilyProperties(physicalDevice_, &queueFamilyPropertiesCount, nullptr);

    auto queueFamilyProperties = new VkQueueFamilyProperties[queueFamilyPropertiesCount];
    vkGetPhysicalDeviceQueueFamilyProperties(physicalDevice_, &queueFamilyPropertiesCount, queueFamilyProperties);

    for(uint32_t i = 0; i < queueFamilyPropertiesCount; i++) {
        if(queueFamilyProperties[i].queueFlags & VK_QUEUE_GRAPHICS_BIT)
            queueIndices.graphics = i;
    }

    delete[] queueFamilyProperties;

    queueIndices.presentation = queueIndices.graphics; //FIXME: this may not always be true!!

    const std::set<uint32_t> queueFamilyIndices = {queueIndices.graphics};
    std::vector<VkDeviceQueueCreateInfo> deviceQueueCreateInfos;
    for(auto queueFamilyIndex : queueFamilyIndices) {
        const float priority = 1.0f;

        VkDeviceQueueCreateInfo queueCreateInfo = {};
        queueCreateInfo.sType = VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO;
        queueCreateInfo.queueFamilyIndex = queueFamilyIndex;
        queueCreateInfo.queueCount = 1;
        queueCreateInfo.pQueuePriorities = &priority;

        deviceQueueCreateInfos.push_back(queueCreateInfo);
    }

    const std::vector<const char*> enabledExtensions = {"VK_KHR_swapchain"};

    VkDeviceCreateInfo deviceCreateInfo = {};
    deviceCreateInfo.sType = VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO;
    deviceCreateInfo.queueCreateInfoCount = deviceQueueCreateInfos.size();
    deviceCreateInfo.pQueueCreateInfos = deviceQueueCreateInfos.data();
    deviceCreateInfo.enabledExtensionCount = enabledExtensions.size();
    deviceCreateInfo.ppEnabledExtensionNames = enabledExtensions.data();

    vkCreateDevice(physicalDevice_, &deviceCreateInfo, nullptr, &device_);

    vkGetDeviceQueue(device_, queueIndices.graphics, 0, &graphicsQueue_);
}

void Renderer::createCommandPool() {
    VkCommandPoolCreateInfo poolCreateInfo = {};
    poolCreateInfo.sType = VK_STRUCTURE_TYPE_COMMAND_POOL_CREATE_INFO;
    poolCreateInfo.queueFamilyIndex = queueIndices.graphics;
    poolCreateInfo.flags = VK_COMMAND_POOL_CREATE_RESET_COMMAND_BUFFER_BIT;

    vkCreateCommandPool(device_, &poolCreateInfo, nullptr, &commandPool_);
}

void Renderer::createPresentationRenderPass() {
    VkAttachmentDescription colorAttachment = {};
    colorAttachment.format = VK_FORMAT_B8G8R8A8_SRGB;
    colorAttachment.samples = VK_SAMPLE_COUNT_1_BIT;
    colorAttachment.loadOp = VK_ATTACHMENT_LOAD_OP_CLEAR;
    colorAttachment.storeOp = VK_ATTACHMENT_STORE_OP_STORE;
    colorAttachment.stencilLoadOp = VK_ATTACHMENT_LOAD_OP_DONT_CARE;
    colorAttachment.stencilStoreOp = VK_ATTACHMENT_STORE_OP_DONT_CARE;
    colorAttachment.initialLayout = VK_IMAGE_LAYOUT_UNDEFINED;
    colorAttachment.finalLayout = VK_IMAGE_LAYOUT_PRESENT_SRC_KHR;

    VkAttachmentReference colorAttachmentRef = {};
    colorAttachmentRef.layout = VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL;

    VkSubpassDescription subpass = {};
    subpass.pipelineBindPoint = VK_PIPELINE_BIND_POINT_GRAPHICS;
    subpass.colorAttachmentCount = 1;
    subpass.pColorAttachments = &colorAttachmentRef;

    std::array<VkSubpassDependency, 1> dependencies;
    dependencies[0].srcSubpass = VK_SUBPASS_EXTERNAL;
    dependencies[0].dstSubpass = 0;
    dependencies[0].srcStageMask = VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT;
    dependencies[0].dstStageMask = VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT;
    dependencies[0].srcAccessMask = VK_ACCESS_MEMORY_READ_BIT;
    dependencies[0].dstAccessMask = VK_ACCESS_COLOR_ATTACHMENT_READ_BIT | VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT;
    dependencies[0].dependencyFlags = VK_DEPENDENCY_BY_REGION_BIT;

    VkRenderPassCreateInfo renderPassInfo = {};
    renderPassInfo.sType = VK_STRUCTURE_TYPE_RENDER_PASS_CREATE_INFO;
    renderPassInfo.attachmentCount = 1;
    renderPassInfo.pAttachments = &colorAttachment;
    renderPassInfo.subpassCount = 1;
    renderPassInfo.pSubpasses = &subpass;
    renderPassInfo.dependencyCount = dependencies.size();
    renderPassInfo.pDependencies = dependencies.data();

    vkCreateRenderPass(device_, &renderPassInfo, nullptr, &presentationRenderPass_);
}

void Renderer::createDescriptorPool() {
    const std::array<VkDescriptorPoolSize, 2> poolSizes = {
        VkDescriptorPoolSize{VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, 50},
        VkDescriptorPoolSize{VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, 50}
    };

    VkDescriptorPoolCreateInfo poolInfo = {};
    poolInfo.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_CREATE_INFO;
    poolInfo.flags = VK_DESCRIPTOR_POOL_CREATE_FREE_DESCRIPTOR_SET_BIT;
    poolInfo.poolSizeCount = poolSizes.size();
    poolInfo.pPoolSizes = poolSizes.data();
    poolInfo.maxSets = 50;

    vkCreateDescriptorPool(device_, &poolInfo, nullptr, &descriptorPool_);
}

void Renderer::createMaterialSetLayout() {
    VkDescriptorSetLayoutBinding albedoSamplerBinding = {};
    albedoSamplerBinding.descriptorCount = 1;
    albedoSamplerBinding.descriptorType = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
    albedoSamplerBinding.stageFlags = VK_SHADER_STAGE_FRAGMENT_BIT;

    VkDescriptorSetLayoutCreateInfo createInfo = {};
    createInfo.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_CREATE_INFO;
    createInfo.bindingCount = 1;
    createInfo.pBindings = &albedoSamplerBinding;

    vkCreateDescriptorSetLayout(device_, &createInfo, nullptr, &materialSetLayout_);
}

void Renderer::createEmptyMaterialSet() {
    VkImageCreateInfo imageCreateInfo = {};
    imageCreateInfo.sType = VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO;
    imageCreateInfo.imageType = VK_IMAGE_TYPE_2D;
    imageCreateInfo.extent.width = 1;
    imageCreateInfo.extent.height = 1;
    imageCreateInfo.extent.depth = 1;
    imageCreateInfo.mipLevels = 1;
    imageCreateInfo.arrayLayers = 1;
    imageCreateInfo.format = VK_FORMAT_R8G8B8A8_UNORM;
    imageCreateInfo.tiling = VK_IMAGE_TILING_OPTIMAL;
    imageCreateInfo.usage = VK_IMAGE_USAGE_TRANSFER_DST_BIT | VK_IMAGE_USAGE_SAMPLED_BIT;
    imageCreateInfo.sharingMode = VK_SHARING_MODE_EXCLUSIVE;
    imageCreateInfo.samples = VK_SAMPLE_COUNT_1_BIT;

    vkCreateImage(device_, &imageCreateInfo, nullptr, &emptyImage_);

    name_object(device_, VK_OBJECT_TYPE_IMAGE, (uint64_t)emptyImage_, "empty image");

    VkMemoryRequirements memRequirements;
    vkGetImageMemoryRequirements(device_, emptyImage_, &memRequirements);

    VkMemoryAllocateInfo allocInfo = {};
    allocInfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
    allocInfo.allocationSize = memRequirements.size;
    allocInfo.memoryTypeIndex = findMemoryType(memRequirements.memoryTypeBits, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT);

    vkAllocateMemory(device_, &allocInfo, nullptr, &emptyMemory_);
    vkBindImageMemory(device_, emptyImage_, emptyMemory_, 0);

    char img[4] = {0};

    uploadImageData(
        emptyImage_,
        1,
        1,
        4,
        img
    );

    VkImageViewCreateInfo createInfo = {};
    createInfo.sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
    createInfo.image = emptyImage_;
    createInfo.viewType = VK_IMAGE_VIEW_TYPE_2D;
    createInfo.format = VK_FORMAT_R8G8B8A8_UNORM;
    createInfo.subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
    createInfo.subresourceRange.levelCount = 1;
    createInfo.subresourceRange.layerCount = 1;

    vkCreateImageView(device_, &createInfo, nullptr, &emptyImageView_);

    VkSamplerCreateInfo samplerInfo = {};
    samplerInfo.sType = VK_STRUCTURE_TYPE_SAMPLER_CREATE_INFO;
    samplerInfo.magFilter = VK_FILTER_LINEAR;
    samplerInfo.minFilter = VK_FILTER_LINEAR;
    samplerInfo.addressModeU = VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_BORDER;
    samplerInfo.addressModeV = VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_BORDER;
    samplerInfo.addressModeW = VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_BORDER;
    samplerInfo.borderColor = VK_BORDER_COLOR_INT_OPAQUE_BLACK;
    samplerInfo.mipmapMode = VK_SAMPLER_MIPMAP_MODE_LINEAR;

    vkCreateSampler(device_, &samplerInfo, nullptr, &emptySampler_);

    VkDescriptorSetAllocateInfo setAllocInfo = {};
    setAllocInfo.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_ALLOCATE_INFO;
    setAllocInfo.descriptorPool = descriptorPool_;
    setAllocInfo.descriptorSetCount = 1;
    setAllocInfo.pSetLayouts = &materialSetLayout_;

    vkAllocateDescriptorSets(device_, &setAllocInfo, &emptyMaterialSet_);

    VkDescriptorImageInfo albedoImageInfo = {};
    albedoImageInfo.imageLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
    albedoImageInfo.imageView = emptyImageView_;
    albedoImageInfo.sampler = emptySampler_;

    VkWriteDescriptorSet albedoDescriptorWrite = {};
    albedoDescriptorWrite.sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
    albedoDescriptorWrite.descriptorCount = 1;
    albedoDescriptorWrite.descriptorType = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
    albedoDescriptorWrite.dstSet = emptyMaterialSet_;
    albedoDescriptorWrite.pImageInfo = &albedoImageInfo;

    vkUpdateDescriptorSets(device_, 1, &albedoDescriptorWrite, 0, nullptr);
}
