#include "smaapass.h"

#include <AreaTex.h>
#include <SearchTex.h>

#include <array>

#include "renderer.h"

SMAAPass::SMAAPass(Renderer& renderer) : renderer(renderer) {
    loadResources();

    setupDescriptorLayouts();
    createOffscreenResources();
    createRenderPasses();
    createPipelines();
}

SMAAPass::~SMAAPass() {
    vkDestroyPipeline(renderer.getDevice(), blendPipeline, nullptr);
    vkDestroyPipeline(renderer.getDevice(), edgePipeline, nullptr);

    vkDestroyPipelineLayout(renderer.getDevice(), edgePipelineLayout, nullptr);
    vkDestroyPipelineLayout(renderer.getDevice(), blendPipelineLayout, nullptr);

    vkDestroyRenderPass(renderer.getDevice(), edgeRenderPass, nullptr);
    vkDestroyRenderPass(renderer.getDevice(), blendRenderPass, nullptr);

    vkDestroySampler(renderer.getDevice(), blendSampler, nullptr);
    vkDestroySampler(renderer.getDevice(), edgeSampler, nullptr);

    vkDestroyDescriptorSetLayout(renderer.getDevice(), blendDescriptorSetLayout, nullptr);
    vkDestroyDescriptorSetLayout(renderer.getDevice(), edgeDescriptorSetLayout, nullptr);

    vkDestroySampler(renderer.getDevice(), searchSampler, nullptr);
    vkDestroySampler(renderer.getDevice(), areaSampler, nullptr);

    vkFreeMemory(renderer.getDevice(), searchMemory, nullptr);
    vkFreeMemory(renderer.getDevice(), areaMemory, nullptr);

    vkDestroyImageView(renderer.getDevice(), searchImageView, nullptr);
    vkDestroyImageView(renderer.getDevice(), areaImageView, nullptr);

    vkDestroyImage(renderer.getDevice(), searchImage, nullptr);
    vkDestroyImage(renderer.getDevice(), areaImage, nullptr);
}

void SMAAPass::render(VkCommandBuffer commandBuffer, RenderTarget* target) {
    vkCmdPipelineBarrier(commandBuffer,
                         VK_PIPELINE_STAGE_FRAGMENT_SHADER_BIT,
                         VK_PIPELINE_STAGE_FRAGMENT_SHADER_BIT,
                         VK_DEPENDENCY_BY_REGION_BIT,
                         0,
                         nullptr,
                         0,
                         nullptr,
                         0,
                         nullptr);

    VkClearValue clearValue = {};
    clearValue.color = {0.0f, 0.0f, 0.0f, 0.0f};

    VkRenderPassBeginInfo renderPassInfo = {};
    renderPassInfo.sType = VK_STRUCTURE_TYPE_RENDER_PASS_BEGIN_INFO;
    renderPassInfo.renderPass = edgeRenderPass;
    renderPassInfo.framebuffer = target->edgeFramebuffers[target->currentResource];
    renderPassInfo.renderArea.extent = target->extent;
    renderPassInfo.clearValueCount = 1;
    renderPassInfo.pClearValues = &clearValue;

    vkCmdBeginRenderPass(commandBuffer, &renderPassInfo, VK_SUBPASS_CONTENTS_INLINE);

    vkCmdBindPipeline(commandBuffer, VK_PIPELINE_BIND_POINT_GRAPHICS, edgePipeline);

    vkCmdBindDescriptorSets(commandBuffer, VK_PIPELINE_BIND_POINT_GRAPHICS, edgePipelineLayout, 0, 1, &target->edgeDescriptorSets[target->currentResource],
                            0, nullptr);

    vkCmdDraw(commandBuffer, 3, 1, 0, 0);

    vkCmdEndRenderPass(commandBuffer);

    renderPassInfo.renderPass = blendRenderPass;
    renderPassInfo.framebuffer = target->blendFramebuffers[target->currentResource];

    vkCmdBeginRenderPass(commandBuffer, &renderPassInfo, VK_SUBPASS_CONTENTS_INLINE);

    // TODO: add a pipeline barrier

    vkCmdBindPipeline(commandBuffer, VK_PIPELINE_BIND_POINT_GRAPHICS, blendPipeline);

    vkCmdBindDescriptorSets(commandBuffer, VK_PIPELINE_BIND_POINT_GRAPHICS, blendPipelineLayout, 0, 1,
                            &target->blendDescriptorSets[target->currentResource], 0, nullptr);

    vkCmdDraw(commandBuffer, 3, 1, 0, 0);

    vkCmdEndRenderPass(commandBuffer);
}

void SMAAPass::createDescriptorSets(RenderTarget* target) {
    // TODO: colmengate these
    VkDescriptorSetAllocateInfo allocInfo = {};
    allocInfo.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_ALLOCATE_INFO;
    allocInfo.descriptorPool = renderer.getDescriptorPool();
    allocInfo.descriptorSetCount = target->numImages;

    // FIXME: lol what
    auto layouts = new VkDescriptorSetLayout[target->numImages];
    for(uint32_t i = 0; i < target->numImages; i++)
        layouts[i] = edgeDescriptorSetLayout;

    allocInfo.pSetLayouts = layouts;

    vkAllocateDescriptorSets(renderer.getDevice(), &allocInfo, target->edgeDescriptorSets);

    for(uint32_t i = 0; i < target->numImages; i++)
        layouts[i] = blendDescriptorSetLayout;

    vkAllocateDescriptorSets(renderer.getDevice(), &allocInfo, target->blendDescriptorSets);

    delete[] layouts;

    for(int i = 0; i < target->numImages; i++) {
        // edge
        {
            VkDescriptorImageInfo imageInfo = {};
            imageInfo.imageLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
            imageInfo.imageView = target->offscreenColorImageViews[i];
            imageInfo.sampler = edgeSampler;

            VkDescriptorImageInfo depthInfo = {};
            depthInfo.imageLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
            depthInfo.imageView = target->offscreenDepthImageViews[i];
            depthInfo.sampler = edgeSampler;

            VkWriteDescriptorSet descriptorWrites[2] = {};
            descriptorWrites[0].sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
            descriptorWrites[0].dstSet = target->edgeDescriptorSets[i];
            descriptorWrites[0].dstBinding = 0;
            descriptorWrites[0].dstArrayElement = 0;
            descriptorWrites[0].descriptorType = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
            descriptorWrites[0].descriptorCount = 1;
            descriptorWrites[0].pImageInfo = &imageInfo;

            descriptorWrites[1].sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
            descriptorWrites[1].dstSet = target->edgeDescriptorSets[i];
            descriptorWrites[1].dstBinding = 1;
            descriptorWrites[1].dstArrayElement = 0;
            descriptorWrites[1].descriptorType = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
            descriptorWrites[1].descriptorCount = 1;
            descriptorWrites[1].pImageInfo = &depthInfo;

            vkUpdateDescriptorSets(renderer.getDevice(), 2, descriptorWrites, 0, nullptr);
        }

        // blend
        {
            VkDescriptorImageInfo edgeInfo = {};
            edgeInfo.imageLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
            edgeInfo.imageView = target->edgeImageViews[i];
            edgeInfo.sampler = edgeSampler;

            VkDescriptorImageInfo areaInfo = {};
            areaInfo.imageLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
            areaInfo.imageView = areaImageView;
            areaInfo.sampler = areaSampler;

            VkDescriptorImageInfo searchInfo = {};
            searchInfo.imageLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
            searchInfo.imageView = searchImageView;
            searchInfo.sampler = searchSampler;

            VkWriteDescriptorSet descriptorWrites[3] = {};
            descriptorWrites[0].sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
            descriptorWrites[0].dstSet = target->blendDescriptorSets[i];
            descriptorWrites[0].dstBinding = 0;
            descriptorWrites[0].dstArrayElement = 0;
            descriptorWrites[0].descriptorType = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
            descriptorWrites[0].descriptorCount = 1;
            descriptorWrites[0].pImageInfo = &edgeInfo;

            descriptorWrites[1].sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
            descriptorWrites[1].dstSet = target->blendDescriptorSets[i];
            descriptorWrites[1].dstBinding = 1;
            descriptorWrites[1].dstArrayElement = 0;
            descriptorWrites[1].descriptorType = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
            descriptorWrites[1].descriptorCount = 1;
            descriptorWrites[1].pImageInfo = &areaInfo;

            descriptorWrites[2].sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
            descriptorWrites[2].dstSet = target->blendDescriptorSets[i];
            descriptorWrites[2].dstBinding = 2;
            descriptorWrites[2].dstArrayElement = 0;
            descriptorWrites[2].descriptorType = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
            descriptorWrites[2].descriptorCount = 1;
            descriptorWrites[2].pImageInfo = &searchInfo;

            vkUpdateDescriptorSets(renderer.getDevice(), 3, descriptorWrites, 0, nullptr);
        }
    }
}

void SMAAPass::loadResources() {
    const auto loadResource = [this](const int size, const int width, const int height, const void* data, VkFormat format, VkImage& image, VkDeviceMemory& memory, VkImageView& view) {
        VkImageCreateInfo imageInfo = {};
        imageInfo.sType = VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO;
        imageInfo.imageType = VK_IMAGE_TYPE_2D;
        imageInfo.extent.width = width;
        imageInfo.extent.height = height;
        imageInfo.extent.depth = 1;
        imageInfo.mipLevels = 1;
        imageInfo.arrayLayers = 1;
        imageInfo.format = format;
        imageInfo.usage = VK_IMAGE_USAGE_TRANSFER_DST_BIT | VK_IMAGE_USAGE_SAMPLED_BIT;
        imageInfo.samples = VK_SAMPLE_COUNT_1_BIT;

        vkCreateImage(renderer.getDevice(), &imageInfo, nullptr, &image);

        VkMemoryRequirements memRequirements = {};
        vkGetImageMemoryRequirements(renderer.getDevice(), image, &memRequirements);

        VkMemoryAllocateInfo allocInfo = {};
        allocInfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
        allocInfo.allocationSize = memRequirements.size;
        allocInfo.memoryTypeIndex = renderer.findMemoryType(memRequirements.memoryTypeBits, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT);

        vkAllocateMemory(renderer.getDevice(), &allocInfo, nullptr, &memory);

        vkBindImageMemory(renderer.getDevice(), image, memory, 0);

        renderer.uploadImageData(
            image,
            width,
            height,
            size,
            data);

        VkImageViewCreateInfo viewInfo = {};
        viewInfo.sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
        viewInfo.image = image;
        viewInfo.viewType = VK_IMAGE_VIEW_TYPE_2D;
        viewInfo.format = imageInfo.format;
        viewInfo.subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
        viewInfo.subresourceRange.levelCount = 1;
        viewInfo.subresourceRange.layerCount = 1;

        vkCreateImageView(renderer.getDevice(), &viewInfo, nullptr, &view);
    };

    loadResource(AREATEX_SIZE, AREATEX_WIDTH, AREATEX_HEIGHT, areaTexBytes, VK_FORMAT_R8G8_UNORM, areaImage, areaMemory, areaImageView);
    loadResource(SEARCHTEX_SIZE, SEARCHTEX_WIDTH, SEARCHTEX_HEIGHT, searchTexBytes, VK_FORMAT_R8_UNORM, searchImage, searchMemory, searchImageView);

    VkSamplerCreateInfo samplerCreateInfo = {};
    samplerCreateInfo.sType = VK_STRUCTURE_TYPE_SAMPLER_CREATE_INFO;
    samplerCreateInfo.magFilter = VK_FILTER_LINEAR;
    samplerCreateInfo.minFilter = VK_FILTER_LINEAR;
    samplerCreateInfo.mipmapMode = VK_SAMPLER_MIPMAP_MODE_LINEAR;
    samplerCreateInfo.addressModeU = VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE;
    samplerCreateInfo.addressModeV = samplerCreateInfo.addressModeU;
    samplerCreateInfo.addressModeW = samplerCreateInfo.addressModeU;
    samplerCreateInfo.mipLodBias = 0.0f;
    samplerCreateInfo.maxAnisotropy = 0;
    samplerCreateInfo.minLod = 0.0f;
    samplerCreateInfo.maxLod = 1.0f;
    samplerCreateInfo.borderColor = VK_BORDER_COLOR_FLOAT_OPAQUE_WHITE;

    vkCreateSampler(renderer.getDevice(), &samplerCreateInfo, nullptr, &areaSampler);

    samplerCreateInfo.magFilter = VK_FILTER_NEAREST;
    samplerCreateInfo.minFilter = VK_FILTER_NEAREST;

    vkCreateSampler(renderer.getDevice(), &samplerCreateInfo, nullptr, &searchSampler);
}

void SMAAPass::setupDescriptorLayouts() {
    // edge
    {
        VkDescriptorSetLayoutBinding imageSamplerLayoutBinding = {};
        imageSamplerLayoutBinding.descriptorType = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
        imageSamplerLayoutBinding.descriptorCount = 1;
        imageSamplerLayoutBinding.stageFlags = VK_SHADER_STAGE_FRAGMENT_BIT;

        VkDescriptorSetLayoutBinding depthSamplerLayoutBinding = {};
        depthSamplerLayoutBinding.binding = 1;
        depthSamplerLayoutBinding.descriptorType = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
        depthSamplerLayoutBinding.descriptorCount = 1;
        depthSamplerLayoutBinding.stageFlags = VK_SHADER_STAGE_FRAGMENT_BIT;

        const std::array<VkDescriptorSetLayoutBinding, 2> bindings = {imageSamplerLayoutBinding, depthSamplerLayoutBinding};

        VkDescriptorSetLayoutCreateInfo layoutInfo = {};
        layoutInfo.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_CREATE_INFO;
        layoutInfo.bindingCount = bindings.size();
        layoutInfo.pBindings = bindings.data();

        vkCreateDescriptorSetLayout(renderer.getDevice(), &layoutInfo, nullptr, &edgeDescriptorSetLayout);
    }

    // blend
    {
        VkDescriptorSetLayoutBinding edgeSamplerLayoutBinding = {};
        edgeSamplerLayoutBinding.descriptorType = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
        edgeSamplerLayoutBinding.descriptorCount = 1;
        edgeSamplerLayoutBinding.stageFlags = VK_SHADER_STAGE_FRAGMENT_BIT;

        VkDescriptorSetLayoutBinding areaSamplerLayoutBinding = {};
        areaSamplerLayoutBinding.binding = 1;
        areaSamplerLayoutBinding.descriptorType = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
        areaSamplerLayoutBinding.descriptorCount = 1;
        areaSamplerLayoutBinding.stageFlags = VK_SHADER_STAGE_FRAGMENT_BIT;

        VkDescriptorSetLayoutBinding searchSamplerLayoutBinding = {};
        searchSamplerLayoutBinding.binding = 2;
        searchSamplerLayoutBinding.descriptorType = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
        searchSamplerLayoutBinding.descriptorCount = 1;
        searchSamplerLayoutBinding.stageFlags = VK_SHADER_STAGE_FRAGMENT_BIT;

        const std::array<VkDescriptorSetLayoutBinding, 3> bindings = {edgeSamplerLayoutBinding, areaSamplerLayoutBinding, searchSamplerLayoutBinding};

        VkDescriptorSetLayoutCreateInfo layoutInfo = {};
        layoutInfo.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_CREATE_INFO;
        layoutInfo.bindingCount = bindings.size();
        layoutInfo.pBindings = bindings.data();

        vkCreateDescriptorSetLayout(renderer.getDevice(), &layoutInfo, nullptr, &blendDescriptorSetLayout);
    }
}

void SMAAPass::createOffscreenResources() {
    const VkFormat colorFormat = VK_FORMAT_R16G16B16A16_SFLOAT;

    VkSamplerCreateInfo samplerCreateInfo = {};
    samplerCreateInfo.sType = VK_STRUCTURE_TYPE_SAMPLER_CREATE_INFO;
    samplerCreateInfo.magFilter = VK_FILTER_LINEAR;
    samplerCreateInfo.minFilter = VK_FILTER_LINEAR;
    samplerCreateInfo.mipmapMode = VK_SAMPLER_MIPMAP_MODE_LINEAR;
    samplerCreateInfo.addressModeU = VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE;
    samplerCreateInfo.addressModeV = samplerCreateInfo.addressModeU;
    samplerCreateInfo.addressModeW = samplerCreateInfo.addressModeU;
    samplerCreateInfo.mipLodBias = 0.0f;
    samplerCreateInfo.maxAnisotropy = 0;
    samplerCreateInfo.minLod = 0.0f;
    samplerCreateInfo.maxLod = 1.0f;
    samplerCreateInfo.borderColor = VK_BORDER_COLOR_FLOAT_OPAQUE_WHITE;

    vkCreateSampler(renderer.getDevice(), &samplerCreateInfo, nullptr, &edgeSampler);
    vkCreateSampler(renderer.getDevice(), &samplerCreateInfo, nullptr, &blendSampler);
}

void SMAAPass::createRenderPasses() {
    VkAttachmentDescription colorAttachment = {};
    colorAttachment.format = VK_FORMAT_R16G16B16A16_SFLOAT;
    colorAttachment.samples = VK_SAMPLE_COUNT_1_BIT;
    colorAttachment.loadOp = VK_ATTACHMENT_LOAD_OP_CLEAR;
    colorAttachment.storeOp = VK_ATTACHMENT_STORE_OP_STORE;
    colorAttachment.stencilLoadOp = VK_ATTACHMENT_LOAD_OP_DONT_CARE;
    colorAttachment.stencilStoreOp = VK_ATTACHMENT_STORE_OP_DONT_CARE;
    colorAttachment.initialLayout = VK_IMAGE_LAYOUT_UNDEFINED;
    colorAttachment.finalLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;

    VkAttachmentReference colorAttachmentRef = {};
    colorAttachmentRef.layout = VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL;

    VkSubpassDescription subpass = {};
    subpass.pipelineBindPoint = VK_PIPELINE_BIND_POINT_GRAPHICS;
    subpass.colorAttachmentCount = 1;
    subpass.pColorAttachments = &colorAttachmentRef;

    VkSubpassDependency dependency = {};
    dependency.srcSubpass = VK_SUBPASS_EXTERNAL;
    dependency.dstSubpass = 0;
    dependency.srcStageMask = VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT;
    dependency.srcAccessMask = 0;
    dependency.dstStageMask = VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT;
    dependency.dstAccessMask = VK_ACCESS_COLOR_ATTACHMENT_READ_BIT | VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT;

    VkRenderPassCreateInfo renderPassInfo = {};
    renderPassInfo.sType = VK_STRUCTURE_TYPE_RENDER_PASS_CREATE_INFO;
    renderPassInfo.attachmentCount = 1;
    renderPassInfo.pAttachments = &colorAttachment;
    renderPassInfo.subpassCount = 1;
    renderPassInfo.pSubpasses = &subpass;
    renderPassInfo.dependencyCount = 1;
    renderPassInfo.pDependencies = &dependency;

    // FIXME: wot
    vkCreateRenderPass(renderer.getDevice(), &renderPassInfo, nullptr, &edgeRenderPass);
    vkCreateRenderPass(renderer.getDevice(), &renderPassInfo, nullptr, &blendRenderPass);
}

void SMAAPass::createPipelines() {
    VkShaderModule edgeVertexShader = renderer.createShader("shaders/edge.vert.spv");
    VkShaderModule edgeFragmentShader = renderer.createShader("shaders/edge.frag.spv");

    VkPipelineShaderStageCreateInfo vertShaderStageInfo = {};
    vertShaderStageInfo.sType = VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO;
    vertShaderStageInfo.stage = VK_SHADER_STAGE_VERTEX_BIT;
    vertShaderStageInfo.module = edgeVertexShader;
    vertShaderStageInfo.pName = "main";

    VkPipelineShaderStageCreateInfo fragShaderStageInfo = {};
    fragShaderStageInfo.sType = VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO;
    fragShaderStageInfo.stage = VK_SHADER_STAGE_FRAGMENT_BIT;
    fragShaderStageInfo.module = edgeFragmentShader;
    fragShaderStageInfo.pName = "main";

    VkPipelineShaderStageCreateInfo shaderStages[] = {vertShaderStageInfo, fragShaderStageInfo};

    VkPipelineVertexInputStateCreateInfo vertexInputInfo = {};
    vertexInputInfo.sType = VK_STRUCTURE_TYPE_PIPELINE_VERTEX_INPUT_STATE_CREATE_INFO;

    VkPipelineInputAssemblyStateCreateInfo inputAssembly = {};
    inputAssembly.sType = VK_STRUCTURE_TYPE_PIPELINE_INPUT_ASSEMBLY_STATE_CREATE_INFO;
    inputAssembly.topology = VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST;
    inputAssembly.primitiveRestartEnable = VK_FALSE;

    VkPipelineViewportStateCreateInfo viewportState = {};
    viewportState.sType = VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_STATE_CREATE_INFO;
    viewportState.viewportCount = 1;
    viewportState.scissorCount = 1;

    VkPipelineRasterizationStateCreateInfo rasterizer = {};
    rasterizer.sType = VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_STATE_CREATE_INFO;
    rasterizer.depthClampEnable = VK_FALSE;
    rasterizer.rasterizerDiscardEnable = VK_FALSE;
    rasterizer.polygonMode = VK_POLYGON_MODE_FILL;
    rasterizer.lineWidth = 1.0f;
    rasterizer.cullMode = VK_CULL_MODE_NONE;
    rasterizer.frontFace = VK_FRONT_FACE_CLOCKWISE;
    rasterizer.depthBiasEnable = VK_FALSE;

    VkPipelineMultisampleStateCreateInfo multisampling = {};
    multisampling.sType = VK_STRUCTURE_TYPE_PIPELINE_MULTISAMPLE_STATE_CREATE_INFO;
    multisampling.sampleShadingEnable = VK_FALSE;
    multisampling.rasterizationSamples = VK_SAMPLE_COUNT_1_BIT;

    VkPipelineColorBlendAttachmentState colorBlendAttachment = {};
    colorBlendAttachment.colorWriteMask =
    VK_COLOR_COMPONENT_R_BIT | VK_COLOR_COMPONENT_G_BIT | VK_COLOR_COMPONENT_B_BIT | VK_COLOR_COMPONENT_A_BIT;
    colorBlendAttachment.blendEnable = VK_FALSE;

    VkPipelineColorBlendStateCreateInfo colorBlending = {};
    colorBlending.sType = VK_STRUCTURE_TYPE_PIPELINE_COLOR_BLEND_STATE_CREATE_INFO;
    colorBlending.attachmentCount = 1;
    colorBlending.pAttachments = &colorBlendAttachment;

    const std::array<VkDynamicState, 2> dynamicStates = {
        VK_DYNAMIC_STATE_VIEWPORT,
        VK_DYNAMIC_STATE_SCISSOR
    };

    VkPipelineDynamicStateCreateInfo dynamicState = {};
    dynamicState.sType = VK_STRUCTURE_TYPE_PIPELINE_DYNAMIC_STATE_CREATE_INFO;
    dynamicState.dynamicStateCount = dynamicStates.size();
    dynamicState.pDynamicStates = dynamicStates.data();

    VkPipelineLayoutCreateInfo pipelineLayoutInfo = {};
    pipelineLayoutInfo.sType = VK_STRUCTURE_TYPE_PIPELINE_LAYOUT_CREATE_INFO;
    pipelineLayoutInfo.setLayoutCount = 1;
    pipelineLayoutInfo.pSetLayouts = &edgeDescriptorSetLayout;

    vkCreatePipelineLayout(renderer.getDevice(), &pipelineLayoutInfo, nullptr, &edgePipelineLayout);

    VkGraphicsPipelineCreateInfo pipelineInfo = {};
    pipelineInfo.sType = VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_CREATE_INFO;
    pipelineInfo.stageCount = 2;
    pipelineInfo.pStages = shaderStages;
    pipelineInfo.pVertexInputState = &vertexInputInfo;
    pipelineInfo.pInputAssemblyState = &inputAssembly;
    pipelineInfo.pViewportState = &viewportState;
    pipelineInfo.pRasterizationState = &rasterizer;
    pipelineInfo.pMultisampleState = &multisampling;
    pipelineInfo.pColorBlendState = &colorBlending;
    pipelineInfo.pDynamicState = &dynamicState;
    pipelineInfo.layout = edgePipelineLayout;
    pipelineInfo.renderPass = edgeRenderPass;

    vkCreateGraphicsPipelines(renderer.getDevice(), nullptr, 1, &pipelineInfo, nullptr, &edgePipeline);

    VkShaderModule blendVertexShader = renderer.createShader("shaders/blend.vert.spv");
    VkShaderModule blendFragmentShader = renderer.createShader("shaders/blend.frag.spv");

    vertShaderStageInfo.module = blendVertexShader;
    fragShaderStageInfo.module = blendFragmentShader;

    VkPipelineShaderStageCreateInfo newShaderStages[] = {vertShaderStageInfo, fragShaderStageInfo};
    pipelineInfo.pStages = newShaderStages;

    pipelineLayoutInfo.pSetLayouts = &blendDescriptorSetLayout;

    vkCreatePipelineLayout(renderer.getDevice(), &pipelineLayoutInfo, nullptr, &blendPipelineLayout);

    pipelineInfo.layout = blendPipelineLayout;
    pipelineInfo.renderPass = blendRenderPass;

    vkCreateGraphicsPipelines(renderer.getDevice(), nullptr, 1, &pipelineInfo, nullptr, &blendPipeline);

    vkDestroyShaderModule(renderer.getDevice(), blendFragmentShader, nullptr);
    vkDestroyShaderModule(renderer.getDevice(), blendVertexShader, nullptr);

    vkDestroyShaderModule(renderer.getDevice(), edgeFragmentShader, nullptr);
    vkDestroyShaderModule(renderer.getDevice(), edgeVertexShader, nullptr);
}
