#include "mainwindow.h"

#include <QCoreApplication>
#include <QMenuBar>
#include <QMenu>
#include <QAction>
#include <QVulkanWindow>
#include <QProcess>
#include <ToolWindowManager.h>

#include "renderwindow.h"
#include "renderer.h"
#include "hierarchy.h"
#include "properties.h"
#include "ecs.h"
#include "worldmanager.h"

MainWindow::MainWindow(Context& context) : context(context) {
    setWindowTitle("Level Editor");
    resize(1280, 720);

    QMenuBar* mainMenuBar = new QMenuBar();
    setMenuBar(mainMenuBar);

    QMenu* fileMenu = new QMenu("File");
    mainMenuBar->addMenu(fileMenu);

    // file
    {
        QAction* newAction = new QAction("New");
        newAction->setIcon(QIcon::fromTheme("document-new"));
        newAction->setShortcut(QKeySequence("CTRL+N"));
        fileMenu->addAction(newAction);

        QAction* openAction = new QAction("Open");
        openAction->setIcon(QIcon::fromTheme("document-open"));
        openAction->setShortcut(QKeySequence("CTRL+O"));
        fileMenu->addAction(openAction);

        QAction* openRecentAction = new QAction("Open Recent...");
        openRecentAction->setIcon(QIcon::fromTheme("document-open-recent"));
        fileMenu->addAction(openRecentAction);

        fileMenu->addSeparator();

        QAction* saveAction = new QAction("Save");
        saveAction->setIcon(QIcon::fromTheme("document-save"));
        saveAction->setShortcut(QKeySequence("CTRL+S"));
        fileMenu->addAction(saveAction);

        QAction* saveAsAction = new QAction("Save As...");
        saveAsAction->setIcon(QIcon::fromTheme("document-save-as"));
        saveAsAction->setShortcut(QKeySequence("CTRL+SHIFT+S"));
        fileMenu->addAction(saveAsAction);

        fileMenu->addSeparator();

        QAction* quitAction = new QAction("Quit");
        connect(quitAction, &QAction::triggered, QCoreApplication::quit);
        quitAction->setIcon(QIcon::fromTheme("application-exit"));
        quitAction->setShortcut(QKeySequence("CTRL+Q"));
        fileMenu->addAction(quitAction);
    }

    QMenu* editMenu = new QMenu("Edit");
    mainMenuBar->addMenu(editMenu);

    // edit
    {
        QAction* undoAction = new QAction("Undo");
        undoAction->setIcon(QIcon::fromTheme("edit-undo"));
        undoAction->setShortcut(QKeySequence("CTRL+Z"));
        editMenu->addAction(undoAction);

        QAction* redoAction = new QAction("Redo");
        redoAction->setIcon(QIcon::fromTheme("edit-redo"));
        redoAction->setShortcut(QKeySequence("CTRL+SHIFT+Z"));
        editMenu->addAction(redoAction);

        editMenu->addSeparator();

        QAction* userPreferenes = new QAction("User Preferences...");
        userPreferenes->setIcon(QIcon::fromTheme("document-properties"));
        editMenu->addAction(userPreferenes);
    }

    QMenu* entityMenu = new QMenu("Entity");
    mainMenuBar->addMenu(entityMenu);

    // entity
    {
        QAction* createEntity = new QAction("Create");
        connect(createEntity, &QAction::triggered, [&context] {
            EntityID entity = ECS::createEntity(worldManager.getCurrentWorld());

            InfoComponent* info = ECS::addComponent<InfoComponent>(entity);
            info->name = "New Entity";

            ECS::addComponent<TransformComponent>(entity);

            emit context.entitiesChanged();
        });
        createEntity->setIcon(QIcon::fromTheme("list-add"));
        entityMenu->addAction(createEntity);
    }

    QMenu* toolsMenu = new QMenu("Tools");
    mainMenuBar->addMenu(toolsMenu);

    // tools
    {
        QAction* materialEditorAction = new QAction("Material Editor");
        connect(materialEditorAction, &QAction::triggered, [] {
            QProcess *process = new QProcess();
            QString file = "tools/materialeditor/MaterialEditor";
            process->start(file);
        });
        materialEditorAction->setIcon(QIcon("data/maticon.png"));
        toolsMenu->addAction(materialEditorAction);
    }

    QMenu* windowMenu = new QMenu("Window");
    mainMenuBar->addMenu(windowMenu);

    // window
    {
        QAction* newWindowAction = new QAction("New Window");
        connect(newWindowAction, &QAction::triggered, [&context] {
            MainWindow* window = new MainWindow(context);
            window->show();
        });
        newWindowAction->setIcon(QIcon::fromTheme("window-new"));
        windowMenu->addAction(newWindowAction);

        QMenu* newViewMenu = new QMenu("New View");
        newViewMenu->setIcon(QIcon::fromTheme("folder-new"));
        windowMenu->addMenu(newViewMenu);

        QAction* addSceneView = new QAction("Scene View");
        connect(addSceneView, &QAction::triggered, this, &MainWindow::addSceneView);
        addSceneView->setIcon(QIcon::fromTheme("view-fullscreen"));
        newViewMenu->addAction(addSceneView);

        QAction* addHierarchy = new QAction("Hierarchy");
        connect(addHierarchy, &QAction::triggered, this, &MainWindow::addHierarchy);
        addHierarchy->setIcon(QIcon::fromTheme("view-sort-ascending"));
        newViewMenu->addAction(addHierarchy);

        QAction* addProperties = new QAction("Properties");
        connect(addProperties, &QAction::triggered, this, &MainWindow::addProperties);
        addProperties->setIcon(QIcon::fromTheme("document-properties"));
        newViewMenu->addAction(addProperties);
    }

    mainMenuBar->addSeparator();

    QMenu* helpMenu = new QMenu("Help");
    mainMenuBar->addMenu(helpMenu);

    instance = new QVulkanInstance();
    instance->setVkInstance(context.renderer->getInstance());
    instance->create();

    manager = new ToolWindowManager();
    setCentralWidget(manager);

    addSceneView();
    addHierarchy();
    addProperties();
}

void MainWindow::addSceneView() {
    RendererWindow* window = new RendererWindow(context);
    window->setVulkanInstance(instance);

    QWidget* wrapper = QWidget::createWindowContainer(window);
    wrapper->setWindowIcon(QIcon::fromTheme("view-fullscreen"));
    wrapper->setWindowTitle("Scene View");

    manager->addToolWindow(wrapper, ToolWindowManager::AreaReference(ToolWindowManager::EmptySpace));
}

void MainWindow::addHierarchy() {
    Hierarchy* hierarchy = new Hierarchy(context);

    manager->addToolWindow(hierarchy, ToolWindowManager::AreaReference(ToolWindowManager::EmptySpace));
}

void MainWindow::addProperties() {
    Properties* properties = new Properties(context);

    manager->addToolWindow(properties, ToolWindowManager::AreaReference(ToolWindowManager::EmptySpace));
}

