#include "mainwindow.h"

#include <QCoreApplication>
#include <QMenuBar>
#include <QMenu>
#include <QAction>
#include <QVulkanWindow>
#include <QHBoxLayout>
#include <QLabel>
#include <QFileDialog>
#include <QJsonDocument>
#include <QJsonObject>
#include <QSettings>

#include "renderwindow.h"
#include "renderer.h"
#include "coloredit.h"
#include "ecs.h"
#include "worldmanager.h"
#include "material.h"
#include "assetmanager.h"

MainWindow::MainWindow(Context& context) : context(context) {
    resize(1280, 720);

    QSettings settings;
    if(!settings.value("recentlyOpened").isNull())
        openedFiles = settings.value("recentlyOpened").toStringList();

    QMenuBar* mainMenuBar = new QMenuBar();
    setMenuBar(mainMenuBar);

    QMenu* fileMenu = new QMenu("File");
    mainMenuBar->addMenu(fileMenu);

    // file
    {
        QAction* newAction = new QAction("New");
        newAction->setIcon(QIcon::fromTheme("document-new"));
        connect(newAction, &QAction::triggered, [this] {
            openMaterial("basic.material");
            currentlyOpenMaterial = "new material";
        });
        newAction->setShortcut(QKeySequence("CTRL+N"));
        fileMenu->addAction(newAction);

        QAction* openAction = new QAction("Open");
        connect(openAction, &QAction::triggered, [this] {
            QString fileName = QFileDialog::getOpenFileName(this, tr("Open File"),
                                                            "../data/",
                                                            tr("Material Files (*.material)"));
            if(!fileName.isEmpty())
                openMaterial(fileName);
        });
        openAction->setIcon(QIcon::fromTheme("document-open"));
        openAction->setShortcut(QKeySequence("CTRL+O"));
        fileMenu->addAction(openAction);

        openRecentAction = new QAction("Open Recent...");
        openRecentAction->setIcon(QIcon::fromTheme("document-open-recent"));
        fileMenu->addAction(openRecentAction);

        fileMenu->addSeparator();

        saveAction = new QAction("Save");
        connect(saveAction, &QAction::triggered, [this] {
            saveMaterial(currentlyOpenMaterial);
        });
        saveAction->setIcon(QIcon::fromTheme("document-save"));
        saveAction->setShortcut(QKeySequence("CTRL+S"));
        fileMenu->addAction(saveAction);

        saveAsAction = new QAction("Save As...");
        connect(saveAsAction, &QAction::triggered, [this] {
            QString fileName = QFileDialog::getSaveFileName(this, tr("Save File"),
                                                            "../data/",
                                                            tr("Material Files (*.material)"));
            if(!fileName.isEmpty())
                saveMaterial(fileName);
        });
        saveAsAction->setIcon(QIcon::fromTheme("document-save-as"));
        saveAsAction->setShortcut(QKeySequence("CTRL+SHIFT+S"));
        fileMenu->addAction(saveAsAction);

        fileMenu->addSeparator();

        QAction* quitAction = new QAction("Quit");
        connect(quitAction, &QAction::triggered, QCoreApplication::quit);
        quitAction->setIcon(QIcon::fromTheme("application-exit"));
        quitAction->setShortcut(QKeySequence("CTRL+Q"));
        fileMenu->addAction(quitAction);
    }

    QWidget* centralWidget = new QWidget();
    setCentralWidget(centralWidget);

    QHBoxLayout* layout = new QHBoxLayout();
    centralWidget->setLayout(layout);

    attributesLayout = new QGridLayout();
    attributesLayout->setAlignment(Qt::AlignTop);
    attributesLayout->setSpacing(0);
    layout->addLayout(attributesLayout);

    QLabel* label = new QLabel("Color");
    attributesLayout->addWidget(label, 0, 0);

    auto meshComponents = ECS::getWorldComponents<MeshComponent>(worldManager.getCurrentWorld());

    auto& [id, mesh] = meshComponents[0];
    meshComponent = mesh;

    instance = new QVulkanInstance();
    instance->setVkInstance(context.renderer->getInstance());
    instance->create();

    RendererWindow* window = new RendererWindow(context);
    window->setVulkanInstance(instance);

    QWidget* wrapper = QWidget::createWindowContainer(window);
    layout->addWidget(wrapper);

    updateControls();
}

MainWindow::~MainWindow() {
    QSettings settings;
    settings.setValue("recentlyOpened", openedFiles);
}

void MainWindow::openMaterial(QString path) {
    currentlyOpenMaterial = path;

    meshComponent->material = assetManager.loadMaterial(path.toStdString());

    ColorEdit* colorEdit = new ColorEdit(meshComponent->material->color);
    attributesLayout->addWidget(colorEdit, 0, 1);

    if(openedFiles.contains(path))
        openedFiles.removeOne(path);

    openedFiles.push_back(path);

    updateControls();
}

void MainWindow::saveMaterial(QString path) {
    MaterialAsset* material = meshComponent->material;

    QString color;
    color = QString::number(material->color.x) + "," + QString::number(material->color.y) + "," + QString::number(material->color.z);

    QJsonObject object {
        {"albedoTexture", material->albedoTexturePath.c_str()},
        {"color", color}
    };

    QJsonDocument document(object);

    QFile jsonFile(path);
    jsonFile.open(QFile::WriteOnly);
    jsonFile.write(document.toJson());

    currentlyOpenMaterial = path;
}

void MainWindow::updateControls() {
    const auto trimPath = [](QString path) {
        return path.mid(path.lastIndexOf('/') + 1);
    };

    if(!currentlyOpenMaterial.isEmpty())
        setWindowTitle("Material Editor - " + trimPath(currentlyOpenMaterial));
    else
        setWindowTitle("Material Editor");

    saveAction->setEnabled(!currentlyOpenMaterial.isEmpty());
    saveAsAction->setEnabled(!currentlyOpenMaterial.isEmpty());

    if(!openedFiles.empty()) {
        QMenu* recentlyOpenedMenu = new QMenu();
        for(auto file : openedFiles) {
            QAction* fileAction = new QAction(trimPath(file));
            connect(fileAction, &QAction::triggered, [this, file] {
                openMaterial(file);
            });
            fileAction->setIcon(QIcon("data/maticon.png"));
            recentlyOpenedMenu->addAction(fileAction);
        }

        recentlyOpenedMenu->addSeparator();

        QAction* clearRecentAction = new QAction("Clear");
        connect(clearRecentAction, &QAction::triggered, [this] {
            openedFiles.clear();
            updateControls();
        });
        clearRecentAction->setIcon(QIcon::fromTheme("edit-clear"));
        recentlyOpenedMenu->addAction(clearRecentAction);

        openRecentAction->setEnabled(true);
        openRecentAction->setMenu(recentlyOpenedMenu);
    } else {
        openRecentAction->setEnabled(false);
    }
}

